UNKNOWN = "Unknown of class name"
FOR = "for"
BILLING_PERIOD = "For billing period"

# DLMS UA 1000-1 Ed 14 7.3.2 Value group B Table 56
INSTANCE = "instance"
CHANNEL = "channel"
UTILITY_SPECIFIC = "utility specific code"
MANUFACTURER_SPECIFIC = "manufacturer specific code"
RESERVED = "reserved"

OVER_VOLTAGE_COUNTER = "over voltage counter"


GLOBAL_METER_RESET_SCRIPT_TABLE = "Global meter reset Script table"
MDI_RESET_END_OF_BILLING_PERIOD_SCRIPT_TABLE = "MDI reset / End of billing period Script table"
TARIFFICATION_SCRIPT_TABLE = "Tariffication Script table"
SET_OUTPUT_SIGNALS_SCRIPT_TABLE = "Set output signals Script table"
DISCONNECT_CONTROL_SCRIPT_TABLE = "Disconnect control Script table"
IMAGE_ACTIVATION_SCRIPT_TABLE = "Image activation Script table"
PUSH_SCRIPT_TABLE = "Push Script table"
RU_STOP_FRAME_SCRIPT_TABLE = "Russian. Stop frame script table"
SPECIAL_DAYS_TABLE = "Special days table"
END_OF_BILLING_PERIOD_SINGLE_ACTION_SCHEDULE = "End of billing period Single action schedule"
DISCONNECT_CONTROL_SINGLE_ACTION_SCHEDULE = "Disconnect control Single action schedule"
IMAGE_ACTIVATION_SINGLE_ACTION_SCHEDULE = "Image activation Single action schedule"
OUTPUT_CONTROL_SINGLE_ACTION_SCHEDULE = "Output control Single action schedule"
PUSH_SINGLE_ACTION_SCHEDULE = "Push Single action schedule"
LOAD_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE = "Load profile control Single action schedule"
M_BUS_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE = "M-Bus profile control Single action schedule"
FUNCTION_CONTROL_SINGLE_ACTION_SCHEDULE = "Function control Single action schedule"
RU_LIMITER_BY_POWER = "RU. Limiter by power"
RU_LIMITER_BY_CURRENT = "RU. Limiter by current"
RU_LIMITER_BY_VOLTAGE = "RU. Limiter by voltage"
RU_LIMITER_BY_MAGNETIC = "RU. Limiter by magnetic"
RU_LIMITER_BY_DIFFERENCE_CURRENT = "RU. Limiter by difference current"
RU_LIMITER_BY_TEMPERATURE = "RU. Limiter by temperature"
RU_IEC_HDLC_SETUP_OPTO = "IEC HDLC Setup Opto"
RU_IEC_HDLC_SETUP_RS_485 = "IEC HDLC Setup RS-485"
RU_IEC_HDLC_SETUP_GSM = "IEC HDLC Setup GSM"
CURRENT_ASSOCIATION = "Current association"
RU_PUBLIC_CLIENT_ASSOCIATION = "Public client association"
RU_METER_READER_ASSOCIATION = "Meter reader association"
RU_UTILITY_SETTING_ASSOCIATION = "Utility setting association"
RU_PUSH = "PUSH"
COSEM_LOGICAL_DEVICE_NAME = "COSEM logical device name"
INVOCATION_COUNTER = "Invocation counter"
RU_SPECIFIC_PASSPORT_DATA_PROFILE = "Russian specific passport data profile"
RU_DEVICE_FACTORY_NUMBER = "Device factory number"
RU_DEVICE_TYPE = "Device Type"
RU_DEVICE_METROLOGICAL_VERSION = "Device metrological version"
RU_PRODUCER_NAME = "Producer name"
RU_DEVICE_RELEASE_DATE = "Device release date"
RU_REMOTE_CONSOLE_SERIAL_NUMBER = "Remote console Serial number"
RU_SPODES_VERSION = "SPODES version"
RU_DEVICE_CONNECTION_SCHEME = "Device Connection scheme"
RU_DEVICE_ID = "device ID"
RU_COUNTER_POINT_DATA = "counter point data"
I_O_CONTROL_SIGNAL_OBJECTS_GLOBAL = "I/O control signal objects, global"
# 6.2.4 Other abstract general purpose OBIS codes
# Program entries Table 60
ACTIVE_FIRMWARE_IDENTIFIER = "Active firmware identifier"
ACTIVE_FIRMWARE_VERSION = "Active firmware version"
ACTIVE_FIRMWARE_SIGNATURE = "Active firmware signature"
# 6.2.13 Register monitor and alarm monitor objects
RU_ALARM_MONITOR_1 = "RU. Load alarm monitor"
# 6.2.19 Standard readout profile objects
GENERAL_DISPLAY_READOUT = "General display readout"
ALTERNATE_DISPLAY_READOUT = "Alternate display readout"
# 6.2.44 Parameter changes and calibration objects
PARAMETER_CHANGES_AND_CALIBRATION_0 = "RU. Correction counter(Configuring)"
# 6.2.48 Status of internal control signals objects
RU_LCD_BACKLIGHT_MODE = "Lcd backlight mode"

POWER_QUALITY_STATUS_DAILY = "Power quality status (profile of daily testimony)"
POWER_QUALITY_STATUS_QUALITY_LOG = "Power quality status (power quality log)"
TIME_OF_OPERATION = "Time of operation"
RU_DURATION_OF_FAILURE_OVERSTRAIN = "Duration of failure/overstrain"
RU_AMBIENT_TEMPERATURE = "Ambient temperature"
EVENTS_RELATED_TO_VOLTAGE = "Events related to voltage"
RU_EVENTS_RELATED_TO_CURRENT = "Events related to current"
RU_EVENTS_RELATED_TO_LOAD_RELAY = "Events related to load relay"
RU_EVENTS_FOR_PROGRAMMING_DEVICE_PARAMETERS = "Events for programming PU parameters"
RU_EXTERNAL_IMPACT_EVENTS = "External impact events"
RU_COMMUNICATION_EVENTS = "Communication events"
RU_ACCESS_CONTROL_EVENTS = "Access control events"
RU_EVENT_CODES_FOR_THE_SELF_DIAGNOSIS_LOG = "Event codes for the self-diagnosis log"
RU_EVENTS_FOR_EXCEEDING_THE_REACTIVE_POWER = "Events for exceeding the reactive power"
RU_CHANNEL_NUMBER_INTERFACE = "channel number (interface)"
COMMUNICATION_ADDRESS = "Communication address"
LOAD_PARAMETERS_PROFILE = "Load parameters profile"
# 6.2.59 Event counter objects
RU_RELAY_TRIGGERING_METER_FOR_OPENING = "RU. Relay triggering meter for opening"
# 6.2.62 Meter tamper event related objects
METER_OPEN_EVENT_COUNTER = "Meter open event counter"
METER_OPEN_EVENT_TIME_STAMP = "Meter open event, time stamp of current event occurrence"
METER_OPEN_EVENT_DURATION = "Meter open event, duration of current event"
METER_OPEN_EVENT_CUMULATIVE_DURATION = "Meter open event, cumulative duration "
TERMINAL_COVER_OPEN_EVENT_COUNTER = "Terminal cover open event counter"
TERMINAL_COVER_OPEN_EVENT_TIME_STAMP = "Terminal cover open event, time stamp of current event occurrence"
TERMINAL_COVER_OPEN_EVENT_DURATION = "Terminal cover open event, duration of current event"
TERMINAL_COVER_OPEN_EVENT_CUMULATIVE = "Terminal cover open event, cumulative duration"
TILT_EVENT_COUNTER = "Tilt event counter"
TILT_EVENT_TIME_STAMP = "Tilt event, time stamp of current event occurrence"
TILT_EVENT_DURATION = "Tilt event, duration of current event"
TILT_EVENT_CUMULATIVE_DURATION = "Tilt event, cumulative duration"
STRONG_DC_MAGNETIC_FIELD_EVENT_COUNTER = "Strong DC magnetic field event counter"
STRONG_DC_MAGNETIC_FIELD_EVENT_TIME_STAMP = "Strong DC magnetic field event, time stamp of current event occurrence"
STRONG_DC_MAGNETIC_FIELD_EVENT_DURATION = "Strong DC magnetic field event, duration of current event"
STRONG_DC_MAGNETIC_FIELD_EVENT_CUMULATIVE_DURATION = "Strong DC magnetic field event, cumulative duration"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_COUNTER = "Supply control switch / valve tamper event counter"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_TIME_STAMP = "Supply control switch / valve tamper event, time stamp of current event occurrence"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_DURATION = "Supply control switch / valve tamper event, duration of current event"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_CUMULATIVE_DURATION = "Supply control switch / valve tamper event, cumulative duration"
METROLOGY_TAMPER_EVENT_COUNTER = "Metrology tamper event counter"
METROLOGY_TAMPER_EVENT_TIME_STAMP = "Metrology tamper event, time stamp of current event occurrence"
METROLOGY_TAMPER_EVENT_DURATION = "Metrology tamper event, duration of current event"
METROLOGY_TAMPER_EVENT_CUMULATIVE_DURATION = "Metrology tamper event, cumulative duration"
COMMUNICATION_TAMPER_EVENT_COUNTER = "Communication tamper event counter"
COMMUNICATION_TAMPER_EVENT_TIME_STAMP = "Communication tamper event, time stamp of current event occurrence"
COMMUNICATION_TAMPER_EVENT_DURATION = "Communication tamper event, duration of current event"
COMMUNICATION_TAMPER_EVENT_CUMULATIVE_DURATION = "Communication tamper event, cumulative duration"
# 6.2.64 Alarm register, Alarm filter and Alarm descriptor objects
RU_ALARM_REGISTER_1 = "RU. Alarm register. Initiative output"
RU_ALARM_FILTER_1 = "RU. Alarm filter. Initiative exit"
RU_ALARM_REGISTER_2 = "RU. Alarm register. Load relay"
RU_ALARM_FILTER_2 = "RU. Alarm filter. Load relay"
# СТО 34.01-5.1-006-2021 Д.11 Журналы событий и захватываемые параметры
RU_VOLTAGE_LOG = "RU. Voltage log"
RU_CURRENT_LOG = "RU. Current log"
RU_COMMUTATION_LOG = "RU. Commutation log"
RU_DATA_CORRECTION_LOG = "RU. Data correction log"
RU_EXTERNAL_IMPACT_LOG = "RU. External impact log"
RU_COMMUNICATION_LOG = "RU. Communication log"
RU_ACCESS_LOG = "RU. Access log"
RU_SELF_DIAGNOSTIC_LOG = "RU. Self diagnostic log"
RU_REACTIVE_POWER_LOG = "RU. Reactive power log"
RU_QUALITY_POWER_LOG = "RU. Quality power log"
RU_STATUS_I_O_LOG = "RU. Status I/O log"
RU_REACTIVE_POWER_LIMIT_LOG = "RU. Reactive power limit log"
RU_TIME_CORRECTION_LOG = "RU. Time correction log"
RU_START_YEAR_LOG = "RU. Start year log"
RU_QUALITY_FOR_CALCULATION_PERIOD_LOG = "RU. Powerline quality for calculation period log"
RU_CONTROL_POWER_LOG = "RU. Control power log"
RU_BATTERY_LOG = "RU. Battery log"
RU_CONTROL_OF_LOAD_RELAY_BLOCKER_LOG = "RU. Control of load relay blocker log"
RU_TEMPERATURE_CONTROL_LOG = "RU. Temperature control log"
RU_VOLTAGE_DEVIATION_LOG = "RU. Voltage deviation log"
RU_LINEAR_VOLTAGE_DEVIATION_LOG = "RU. Linear voltage deviation log"
RU_ABNORMAL_NETWORK_SITUATION_LOG = "RU. Abnormal network situation log"
RU_VOLTAGE_INTERRUPTION_LOG = "RU. Over interruption log"
RU_OVER_VOLTAGE_LOG = "RU. Over voltage log"
#
ITE_FIRMWARE_DESCRIPTOR = "ITE. Firmware descriptor"
ITE_MAGNETIC_SENSOR_STATUS = "ITE. Magnetic sensor status"
ITE_DISCRETE_OUTPUTS = "ITE. Discrete outputs"
ITE_SETTING_OF_RELAY_INCLUSION_PER_DAY = "ITE. Setting of relay inclusion per day"
ITE_SETTINGS_MESSAGES = "ITE. Settings messages"
ITE_CORE_REGISTERS = "ITE. Core registers"
ITE_BLE_ID = "ITE. BLUETOOTHLE ID"
ITE_BITMAP = "ITE. BITMAP loader"
ITE_ICCID = "ITE. ICCID"
RU_BODY_OPENING_STATE = "RU. Current condition of the opening sensor of the body"
RU_TERMINALS_COVER_OPENING_STATE = "RU. Current condition of the opening sensor of the terminal covers"
RU_MAGNETIC_FIELD_STATE = "RU. Current state of the magnetic field sensor"
RU_HF_FIELD_STATE = "RU. Current state of HF field"
RU_ELECTRONIC_SEALS_FIXED_STATE_OF_EVENTS = "RU. Fixed state of events of electronic seals"
RU_PRESSING_ELECTRONIC_SEALS = "RU. Pressing electronic seals"
RU_CLEAR_OF_ELECTRONIC_SEALS_FIXED_STATE = "RU. Cleaning recorded events (magnet and HF field)"
RU_FIRST_OPENING_TIME_OF_BODY = "RU. The first opening time of the electronic seal of the case"
RU_FIRST_OPENING_TIME_OF_TERMINALS_COVER = "RU. The first opening time of the electronic seal of the terminal covers"
# 7.4. Abstract object
# 7.4.1 Table 60 General and service entry object - Abstract
RU_RELAY_LOAD_ARBITRATOR = "RU. Relay load arbitrator"
# Communication port log parameters
ITE_USED_COMMUNICATION_INTERFACES = "ITE. Used communication interfaces"
# Consumer messages
LOCAL_CONSUMER_MESSAGE = "Consumer message via local consumer information port"
DISPLAY_CONSUMER_MESSAGE = "Consumer message via the meter display and / or via consumer information port"
# 7.4.5 Data profile objects – Abstract
GSM_DIAGNOSTIC_PROFILE = "GSM diagnostic profile"
# electricity
# 7.5.1 Value group C codes – Electricity
ACTIVE_POWER_PLUS = "Active power+ (QI+QIV)"
ACTIVE_POWER_MINUS = "Active power- (QII+QIII)"
REACTIVE_POWER_PLUS = "Reactive power+ (QI+QII)"
REACTIVE_POWER_MINUS = "Reactive power- (QIII+QIV)"
REACTIVE_POWER_QI = "Reactive power QI"
REACTIVE_POWER_QII = "Reactive power QII"
REACTIVE_POWER_QIII = "Reactive power QIII"
REACTIVE_POWER_QIV = "Reactive power QIV"
APPARENT_POWER_PLUS = "Apparent power+ (QI+QIV)"
APPARENT_POWER_MINUS = "Apparent power- (QII+QIII)"
CURRENT = "Current"
VOLTAGE = "Voltage"
POWER_FACTOR = "Power factor"
SUPPLY_FREQUENCY = "Supply frequency"
ACTIVE_POWER_ABS_PLUS = "Active power (abs(QI+QIV)+(abs(QII+QIII))"
ACTIVE_POWER_ABS_MINUS = "Active power (abs(QI+QIV)-abs(QII+QIII))"
ACTIVE_POWER_QI = "Active power QI"
ACTIVE_POWER_QII = "Active power QII"
ACTIVE_POWER_QIII = "Active power QIII"
ACTIVE_POWER_QIV = "Active power QIV"
REACTIVE_FACTOR = "Reactive factor"
CUMULATIVE = "Cumulative"
ANY_PHASE = "Any phase"
ALL_PHASE = "All phase"
L1 = "L1"
L2 = "L2"
L3 = "L3"
L1_L2 = "L1-L2"
L2_L3 = "L2-L3"
L3_L1 = "L3-L1"
CUMULATIVE_AMPERE_SQUARED_HOURS = "Cumulative Ampere-squared hours (QI+QII+QIII+QIV)"
CUMULATIVE_VOLT_SQUARED_HOURS = "Cumulative Volt-squared hours (QI+QII+QIII+QIV)"
L0_CURRENT_NEUTRAL = "L0 current (neutral)"
RU_DIFFERENTIAL_CURRENT = "RU. Differential current"
PERCENT = "Percent"
RU_LINEAR_VOLTAGE = "RU. Linear voltage"
RU_POSITIVE_DEVIATION = "RU. Positive deviation"
RU_NEGATIVE_DEVIATION = "RU. Negative deviation"
POWER_REACTIVE_FACTOR = "Power reactive factor"
VOLTAGE_ASYMMETRY_COEFFICIENT = "RU. Voltage asymmetry coefficient"
# 7.5.2 Value group D codes – Electricity
# 7.5.2.1 Processing of measurement values
BILLING_PERIOD_AVERAGE_SINCE_LAST_RESET = "Billing period average (since last reset)"
CUMULATIVE_MINIMUM_1 = "Cumulative minimum 1"
CUMULATIVE_MAXIMUM_1 = "Cumulative maximum 1"
MINIMUM_1 = "Minimum 1"
CURRENT_AVERAGE_1 = "Current average 1"
LAST_AVERAGE_1 = "Last average 1"
MAXIMUM_1 = "Maximum 1"
INSTANTANEOUS_VALUE = "Instantaneous value"
TIME_INTEGRAL_1 = "Time integral 1"
TIME_INTEGRAL_2 = "Time integral 2"
TIME_INTEGRAL_3 = "Time integral 3"
CUMULATIVE_MINIMUM_2 = "Cumulative minimum 2"
CUMULATIVE_MAXIMUM_2 = "Cumulative maximum 2"
MINIMUM_2 = "Minimum 2"
CURRENT_AVERAGE_2 = "Current average 2"
LAST_AVERAGE_2 = "Last average 2"
MAXIMUM_2 = "Maximum 2"
TIME_INTEGRAL_7 = "Time integral 7"
TIME_INTEGRAL_8 = "Time integral 8"
TIME_INTEGRAL_9 = "Time integral 9"
TIME_INTEGRAL_10 = "Time integral 10"
CUMULATIVE_MINIMUM_3 = "Cumulative minimum 3"
CUMULATIVE_MAXIMUM_3 = "Cumulative maximum 3"
MINIMUM_3 = "Minimum 3"
CURRENT_AVERAGE_3 = "Current average 3"
LAST_AVERAGE_3 = "Last average 3"
MAXIMUM_3 = "Maximum 3"
CURRENT_AVERAGE_5 = "Current average 5"
CURRENT_AVERAGE_6 = "Current average 6"
TIME_INTEGRAL_5 = "Time integral 5"
TIME_INTEGRAL_6 = "Time integral 6"
UNDER_LIMIT_THRESHOLD = "Under limit threshold"
UNDER_LIMIT_OCCURRENCE_COUNTER = "Under limit occurrence counter"
UNDER_LIMIT_DURATION = "Under limit duration"
UNDER_LIMIT_MAGNITUDE = "Under limit magnitude"
OVER_LIMIT_THRESHOLD = "Over limit threshold"
OVER_LIMIT_OCCURRENCE_COUNTER = "Over limit occurrence counter"
OVER_LIMIT_DURATION = "Over limit duration"
OVER_LIMIT_MAGNITUDE = "Over limit magnitude"
MISSING_THRESHOLD = "Missing threshold"
MISSING_OCCURRENCE_COUNTER = "Missing occurrence counter"
MISSING_DURATION = "Missing duration"
MISSING_MAGNITUDE = "Missing magnitude"
TIME_THRESHOLD_FOR_UNDER_LIMIT = "Time threshold for under limit"
TIME_THRESHOLD_FOR_OVER_LIMIT = "Time threshold for over limit"
TIME_THRESHOLD_FOR_MISSING_MAGNITUDE = "Time threshold for missing magnitude"
CONTRACTED_VALUE = "Contracted value"
AVERAGE_VALUE_FOR_RECORDING_INTERVAL_1 = "Average value for recording interval 1"
AVERAGE_VALUE_FOR_RECORDING_INTERVAL_2 = "Average value for recording interval 2"
MINIMUM_FOR_RECORDING_INTERVAL_1 = "Minimum for recording interval 1"
MINIMUM_FOR_RECORDING_INTERVAL_2 = "Minimum for recording interval 2"
MAXIMUM_FOR_RECORDING_INTERVAL_1 = "Maximum for recording interval 1"
MAXIMUM_FOR_RECORDING_INTERVAL_2 = "Maximum for recording interval 2"
TEST_AVERAGE = "Test average"
CURRENT_AVERAGE_4_FOR_HARMONICS_MEASUREMENT = "Current average 4 for harmonics measurement"
TIME_INTEGRAL_4 = "Time integral 4"
RU_TOTAL_DEVIATION_TIME = "RU. Total deviation time"
# 7.5.3.2 Tariff rates
TOTAL = "Total"
RATE = "Rate"
MANUFACTURER_SPECIFIC_CODES = "Manufacturer specific codes"

# 7.5.3.3 Harmonics
TOTAL_FUND_ALL = "Total (fundamental + all harmonics)"
HARMONIC = "harmonic"
THD = "Total Harmonic Distortion(THD)"
TDD = "Total Demand Distortion (TDD)"
ALL_HARMONICS = "All harmonics"
ALL_HARMONICS_TO_NOMINAL_VALUE_RATIO = "All harmonics to nominal value ratio"

# 7.5.3.4 Phase angles
ANGLE_FROM = "Angle from"
TO = "to"
U_L1 = "U(L1)"
U_L2 = "U(L2)"
U_L3 = "U(L3)"
ERROR = "error"
I_L1 = "I(L1)"
I_L2 = "I(L2)"
I_L3 = "I(L3)"
I_L0 = "I(L0)"

# 7.5.5.1 Table 72 General and service entry objects – Electricity
TRANSFORMER_RATIO = "transformer ratio"
COMPLETE_COMBINED_ELECTRICITY_ID = "Complete combined Electricity ID"
# Program entries
""" init in abstract Program entries"""
# Output pulse values or constants
ACTIVE_ENERGY_OUTPUT_PULSE = "Active energy, output pulse"
REACTIVE_ENERGY_OUTPUT_PULSE = "Reactive energy, output pulse"
# Ratios
TRANSFORMER_RATIO_CURRENT = F"{TRANSFORMER_RATIO} - current (numerator)"
TRANSFORMER_RATIO_VOLTAGE = F"{TRANSFORMER_RATIO} - voltage (numerator)"
# Nominal values
NOMINAL_VOLTAGE = "Nominal voltage"
NOMINAL_CURRENT = "Nominal current"
NOMINAL_FREQUENCY = "Nominal frequency"
MAXIMUM_CURRENT = "Maximum current"
REFERENCE_VOLTAGE_FOR_POWER_QUALITY_MEASUREMENT = "Reference voltage for power quality measurement"
REFERENCED_VOLTAGE_FOR_AUX_POWER_SUPPLY = "Reference voltage for aux. power supply"
# Measurement period- / recording interval- / billing period duration
RECORDING_INTERVAL_1_FOR_LOAD_PROFILE = "Recording interval 1, for load profile"
RECORDING_INTERVAL_2_FOR_LOAD_PROFILE = "Recording interval 2, for load profile"
# Coefficients
TRANSFORMER_MAGNETIC_LOSSES = "Transformer magnetic losses"
TRANSFORMER_IRON_LOSSES = "Transformer iron losses"
LINE_RESISTANCE_LOSSES = "Line resistance losses"
LINE_REACTANCE_LOSSES = "Line reactance losses"
# 7.5.5.3 List objects – Electricity
RU_MONTHLY_PROFILE = "RU. Monthly profile"
RU_DAILY_PROFILE = "RU. Daily profile"

# 7.5.5.4 Data profile objects – Electricity
RU_LOAD_PROFILE = "RU. Load profile"


# Country specific identifiers. Russian profiles
COUNTRY_SPECIFIC_IDENTIFIER = "Country specific identifier"
RU_PROFILE_OF_CURRENT_VALUES = "RU. Profile of current values"
RU_SCALE_PROFILE_FOR_THE_MAGAZINE_OF_MONTHLY_INDICATIONS = "RU. Scale profile for the magazine of monthly indications"
RU_SCALE_PROFILE_FOR_A_JOURNAL_OF_DAILY_INDICATION = "RU. Scale profile for a journal of daily indication"
RU_SCALE_PROFILE_FOR_CURRENT_FRAMES_OF_CURRENT_VALUES = "RU. Scale profile for current-frames of current values"
RU_SCALE_PROFILE_FOR_LOAD_PROFILES = "RU. Scale profile for load profiles"
RU_TELEMECHANICS_PROFILE_FOR_TELEVISION_MEASUREMENTS = "RU. Telemechanics profile for television measurements"
RU_TELEMECHANICS_PROFILE_OF_TELEVISION_SIGNALING = "RU. Telemechanics profile of television signaling"
RU_CHANGE_LIMIT_LEVEL = "RU. Change limit level"
# ITE
ITE_CALIBRATION_STATUS = "ITE. Calibration status"
ITE_CALIBRATION_APPARENT_POWER = "ITE. Calibration. Apparent power"
ITE_CALIBRATION_ACTIVE_POWER = "ITE. Calibration. Active power"
ITE_CALIBRATION_REACTIVE_POWER = "ITE. Calibration. Reactive power"
ITE_CALIBRATION_VOLTAGE = "ITE. Calibration. Voltage"
ITE_CALIBRATION_CURRENT = "ITE. Calibration. Current"
ITE_CALIBRATION_ANGLE = "ITE. Calibration. Angle"
ITE_DISPLAY_SETTING_1 = "ITE. Display setting #1"
ITE_DISPLAY_SETTING_2 = "ITE. Display setting #2"
ITE_CLOCK_OFFSET_SETTING = "ITE. Clock offset setting"
ITE_FACTORY_SETTING_10 = "ITE Factory setting #10"
ITE_FACTORY_SETTING_11 = "ITE Factory setting #11"
ITE_FACTORY_SETTING_12 = "ITE Factory setting #12"
ITE_FACTORY_SETTING_13 = "ITE Factory setting #13"
ITE_FACTORY_SETTING_14 = "ITE Factory setting #14"
ITE_FACTORY_SETTING_15 = "ITE Factory setting #15"
ITE_FACTORY_SETTING_16 = "ITE Factory setting #16"
ITE_FACTORY_SETTING_17 = "ITE Factory setting #17"
ITE_FACTORY_SETTING_18 = "ITE Factory setting #18"
ITE_FACTORY_SETTING_19 = "ITE Factory setting #19"

# RU. Simple OBIS. СТО 34.01-5.1-006-2021.
RU_TOTAL_VOLTAGE_DEVIATION_TIME_FOR_CALCULATED_PERIOD = "RU. Total voltage deviation time for calculated period"
RU_LOAD_LOCK_STATUS = "RU. Load lock status"  # 13.5.6
# 13.9. Контроль чередования фаз
SPODES3_PHASE_ALTERNATING_CONTROL = "RU. SPODES3. Phase alternating control"
SPODES3_NOT_METROLOGICAL_FIRMWARE_ID = "RU. SPODES3. Not metrological firmware id"


# 7.2.4 Summary of rules for manufacturer, utility, consortia and country specific codes. Table 54
# KPZ - Kurgansky Priborostroitelny Zavod
KPZ_INTERFACE_LIST = "KPZ. Interface list"



from .types import cst
from .cosem_interface_classes import cosem_interface_class as ic, collection
from . import ITE_exceptions as exc
from .cosem_interface_classes.overview import ClassID


def get_attr_index(obj: ic.COSEMInterfaceClasses) -> list[int]:
    """ return attribute indexes for reading keep configuration for SPODES3 only"""
    match obj.CLASS_ID, obj.logical_name:
        case ClassID.DATA, _:                                                         return [2]
        case ClassID.REGISTER, _:                                                     return [2, 3]
        case ClassID.EXT_REGISTER, _:                                                 return [3]
        case ClassID.PROFILE_GENERIC, cst.LogicalName(1, _, 94, 7, 1 | 2 | 3 | 4):    return [6, 3, 2, 4, 5, 8]
        case ClassID.PROFILE_GENERIC, _:                                              return [6, 3, 4, 5, 8]
        case ClassID.CLOCK, _:                                                        return [8, 9]
        case ClassID.SCRIPT_TABLE, _:                                                 return [2]
        case ClassID.SCHEDULE, _:                                                     return [2]
        case ClassID.SPECIAL_DAYS_TABLE, _:                                           return []
        case ClassID.ACTIVITY_CALENDAR, _:                                            return []
        case ClassID.SINGLE_ACTION_SCHEDULE, _:                                       return [2, 3, 4]
        case ClassID.ASSOCIATION_LN_CLASS, cst.LogicalName(0, _, 40, 0, 0):           return []
        case ClassID.ASSOCIATION_LN_CLASS, _:                                         return [4, 5, 7]
        case ClassID.IEC_HDLC_SETUP, _:                                               return [2, 3, 4, 5, 6, 7]
        case ClassID.DISCONNECT_CONTROL, _:                                           return [3, 4]
        case ClassID.LIMITER, _:                                                      return [2, 11]
        case ClassID.MODEM_CONFIGURATION, _:                                          return [2]
        case ClassID.IMAGE_TRANSFER, _:                                               return [2]
        case ClassID.GPRS_MODEM_SETUP, _:                                             return [2]
        case ClassID.GSM_DIAGNOSTIC, _:                                               return []
        case ClassID.CLIENT_SETUP, _:                                                 return []  # not need only for client
        case ClassID.TCP_UDP_SETUP, _:                                                return [2, 3, 4, 5, 6]
        case ClassID.IPV4_SETUP, _:                                                   return []
        case ClassID.ARBITRATOR, _:                                                   return [2]
        case ClassID.SECURITY_SETUP, _:                                               return [2, 3, 5]
        case ClassID.REGISTER_MONITOR, _:                                             return [3, 2, 4]
        case ClassID.DEMAND_REGISTER, _:                                              return [4, 2, 3, 5, 6, 7, 8, 9]
        case _: raise exc.NoObject(F"Configuring. Not found {obj} with {obj.CLASS_ID} for read attributes")


empty_dict = dict()


def get_saved_parameters(obj: ic.COSEMInterfaceClasses) -> dict[int, int]:
    """ return attribute indexes for saved keep configuration dictionary(attr_index: 0-for value 1-for type, ...)"""
    ln = obj.logical_name
    match obj.CLASS_ID, obj.logical_name:
        case collection.Data.CLASS_ID, cst.LogicalName(0, 0, 96, 1, 1 | 3 | 6 | 8) | cst.LogicalName(1, 0, 0, 8, 4):  return {2: 0}
        case collection.Data.CLASS_ID, cst.LogicalName(0, 0, 96, 11, 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8): return empty_dict
        case collection.Data.CLASS_ID, _:                                                        return {2: 1}
        case collection.Register.CLASS_ID, cst.LogicalName(1, 0, 0, 6, 0 | 1 | 2 | 3 | 4):       return {2: 1, 3: 1}
        case collection.Register.CLASS_ID, _:                                                    return {2: 1, 3: 0}
        case collection.ExtendedRegister.CLASS_ID, _:                                            return {2: 1, 3: 0}
        case collection.ProfileGenericVer1.CLASS_ID, cst.LogicalName(1, _, 94, 7, 1 | 2 | 3 | 4): return {6: 0, 3: 0, 2: 0, 4: 0, 5: 0, 8: 0}
        case collection.ProfileGenericVer1.CLASS_ID, _:                                          return {6: 0, 3: 0, 4: 0, 5: 0, 8: 0}
        case collection.Clock.CLASS_ID, _:                                                       return {8: 0, 9: 0}
        case collection.ScriptTable.CLASS_ID, _:                                                 return {2: 0}
        case collection.Schedule.CLASS_ID, _:                                                    return {2: 0}
        case collection.SpecialDaysTable.CLASS_ID, _:                                            return empty_dict
        case collection.ActivityCalendar.CLASS_ID, _:                                            return empty_dict
        case collection.SingleActionSchedule.CLASS_ID, _:                                        return {2: 0, 3: 0, 4: 0}
        case collection.AssociationLNVer0.CLASS_ID, cst.LogicalName(0, 0, 40, 0, 0):             return {3: 0}
        case collection.AssociationLNVer0.CLASS_ID, cst.LogicalName(0, 0, 40, 0, 1):             return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 9: 0}
        case collection.AssociationLNVer0.CLASS_ID, _:                                           return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 9: 0}
        case collection.IECHDLCSetupVer1.CLASS_ID, _:                                            return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0}
        case collection.DisconnectControl.CLASS_ID, _:                                           return {3: 0, 4: 0}
        case collection.Limiter.CLASS_ID, _:                                                     return {2: 0, 11: 0}
        case collection.PSTNModemConfiguration.CLASS_ID, _:                                      return {2: 0}
        case collection.ImageTransfer.CLASS_ID, _:                                               return {2: 0}
        case collection.GPRSModemSetup.CLASS_ID, _:                                              return {2: 0}
        case collection.GSMDiagnosticVer0.CLASS_ID, _:                                           return empty_dict
        case collection.ClientSetup.CLASS_ID, _:                                                 return empty_dict  # not need only for client
        case collection.TCPUDPSetup.CLASS_ID, _:                                                 return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0}
        case collection.IPv4Setup.CLASS_ID, _:                                                   return empty_dict
        case collection.Arbitrator.CLASS_ID, _:                                                  return {2: 0}
        case collection.SecuritySetupVer0.CLASS_ID, _:                                           return {2: 0, 3: 0, 5: 0}
        case collection.RegisterMonitor.CLASS_ID, _:                                             return {3: 0, 2: 0, 4: 0}
        case ClassID.DEMAND_REGISTER, _:                                                         return {2: 1, 3: 1, 4: 0, 5: 1, 8: 0, 9: 0}
        case _: raise exc.NoObject(F'Save configure. Not found {obj} with {obj.CLASS_ID} for read attributes')


if __name__ == '__main__':
    a = collection.AssociationLNVer0('0.0.1.0.0.255')
    print(a)
    b = get_attr_index(a)
    print(b)


""" The OBIS identification system serves as a basis for the COSEM logical names. The system of naming COSEM objects is defined in the basic
principles (see Clause 4 EN 62056-62:2007), the identification of real data items is specified in IEC 62056-61. The following clauses define the
usage of those definitions in the COSEM environment. All codes, which are not explicitly listed, but outside the manufacturer specific range are
reserved for future use."""
from __future__ import annotations
from functools import lru_cache
from . import settings
from .types import cosem_service_types as cst


match settings.get_current_language():
    case settings.Language.ENGLISH:     from .Values.EN import relation_to_obis_names as rn, class_names as cn
    case settings.Language.RUSSIAN:     from .Values.RU import relation_to_obis_names as rn, class_names as cn


def get_obj_names(electric_obj: int) -> str:
    """ corresponding with DLMS UA 1000-1 Ed. 14 7.5.1 Table 65. Value group C codes – Electricity. Range: 1..80, RU: 124..126 """
    match electric_obj:
        case 1 | 21 | 41 | 61:  return rn.ACTIVE_POWER_PLUS
        case 2 | 22 | 42 | 62:  return rn.ACTIVE_POWER_MINUS
        case 3 | 23 | 43 | 63:  return rn.REACTIVE_POWER_PLUS
        case 4 | 24 | 44 | 64:  return rn.REACTIVE_POWER_MINUS
        case 5 | 25 | 45 | 65:  return rn.REACTIVE_POWER_QI
        case 6 | 26 | 46 | 66:  return rn.REACTIVE_POWER_QII
        case 7 | 27 | 47 | 67:  return rn.REACTIVE_POWER_QIII
        case 8 | 28 | 48 | 68:  return rn.REACTIVE_POWER_QIV
        case 9 | 29 | 49 | 69:  return rn.APPARENT_POWER_PLUS
        case 10 | 30 | 50 | 70: return rn.APPARENT_POWER_MINUS
        case 11 | 31 | 51 | 71: return rn.CURRENT
        case 12 | 32 | 52 | 72: return rn.VOLTAGE
        case 13 | 33 | 53 | 73: return rn.POWER_FACTOR
        case 14 | 34 | 54 | 74: return rn.SUPPLY_FREQUENCY
        case 15 | 35 | 55 | 75: return rn.ACTIVE_POWER_ABS_PLUS
        case 16 | 36 | 56 | 76: return rn.ACTIVE_POWER_ABS_MINUS
        case 17 | 37 | 57 | 77: return rn.ACTIVE_POWER_QI
        case 18 | 38 | 58 | 78: return rn.ACTIVE_POWER_QII
        case 19 | 39 | 59 | 79: return rn.ACTIVE_POWER_QIII
        case 20 | 40 | 60 | 80:  return rn.ACTIVE_POWER_QIV
        case 88: return rn.CUMULATIVE_AMPERE_SQUARED_HOURS
        case 89: return rn.CUMULATIVE_VOLT_SQUARED_HOURS
        case 124: return F"{rn.RU_LINEAR_VOLTAGE} {rn.L1_L2}"
        case 125: return F"{rn.RU_LINEAR_VOLTAGE} {rn.L2_L3}"
        case 126: return F"{rn.RU_LINEAR_VOLTAGE} {rn.L3_L1}"
        case 128: return F"{rn.POWER_REACTIVE_FACTOR} {rn.L1}"
        case 129: return F"{rn.POWER_REACTIVE_FACTOR} {rn.L2}"
        case 130: return F"{rn.POWER_REACTIVE_FACTOR} {rn.L3}"
        case 131: return F"{rn.POWER_REACTIVE_FACTOR} {rn.ALL_PHASE}"
        case 133: return F"{rn.VOLTAGE_ASYMMETRY_COEFFICIENT}"
        case 147: return F"{rn.RU_TOTAL_DEVIATION_TIME} {rn.FOR} {rn.VOLTAGE}"
        case _:   raise ValueError(F"getting electrical name by C group with unknown: C={electric_obj}")


def get_harmonics_name(classification: int) -> str:
    """ See DLMS UA 1000-1 Ed.13 7.5.3.3 Harmonics """
    match classification:
        case 0:                               return rn.TOTAL_FUND_ALL
        case 124:                             return rn.THD
        case 125:                             return rn.TDD
        case 126:                             return rn.ALL_HARMONICS
        case 127:                             return rn.ALL_HARMONICS_TO_NOMINAL_VALUE_RATIO
        case _ if 120 >= classification >= 1: return F"{rn.HARMONIC} {classification}"
        case _: raise ValueError(F"for get harmonics unknown {classification=}")


def get_processing_names(d: int) -> str:
    match d:
        case 0: return rn.BILLING_PERIOD_AVERAGE_SINCE_LAST_RESET
        case 1: return rn.CUMULATIVE_MINIMUM_1
        case 2: return rn.CUMULATIVE_MAXIMUM_1
        case 3: return rn.MINIMUM_1
        case 4: return rn.CURRENT_AVERAGE_1
        case 5: return rn.LAST_AVERAGE_1
        case 6: return rn.MAXIMUM_1
        case 7: return rn.INSTANTANEOUS_VALUE
        case 8: return rn.TIME_INTEGRAL_1
        case 9: return rn.TIME_INTEGRAL_2
        case 10: return rn.TIME_INTEGRAL_3
        case 11: return rn.CUMULATIVE_MINIMUM_2
        case 12: return rn.CUMULATIVE_MAXIMUM_2
        case 13: return rn.MINIMUM_2
        case 14: return rn.CURRENT_AVERAGE_2
        case 15: return rn.LAST_AVERAGE_2
        case 16: return rn.MAXIMUM_2
        case 17: return rn.TIME_INTEGRAL_7
        case 18: return rn.TIME_INTEGRAL_8
        case 19: return rn.TIME_INTEGRAL_9
        case 20: return rn.TIME_INTEGRAL_10
        case 21: return rn.CUMULATIVE_MINIMUM_3
        case 22: return rn.CUMULATIVE_MAXIMUM_3
        case 23: return rn.MINIMUM_3
        case 24: return rn.CURRENT_AVERAGE_3
        case 25: return rn.LAST_AVERAGE_3
        case 26: return rn.MAXIMUM_3
        case 27: return rn.CURRENT_AVERAGE_5
        case 28: return rn.CURRENT_AVERAGE_6
        case 29: return rn.TIME_INTEGRAL_5
        case 30: return rn.TIME_INTEGRAL_6
        case 31: return rn.UNDER_LIMIT_THRESHOLD
        case 32: return rn.UNDER_LIMIT_OCCURRENCE_COUNTER
        case 33: return rn.UNDER_LIMIT_DURATION
        case 34: return rn.UNDER_LIMIT_MAGNITUDE
        case 35: return rn.OVER_LIMIT_THRESHOLD
        case 36: return rn.OVER_LIMIT_OCCURRENCE_COUNTER
        case 37: return rn.OVER_LIMIT_DURATION
        case 38: return rn.OVER_LIMIT_MAGNITUDE
        case 39: return rn.MISSING_THRESHOLD
        case 40: return rn.MISSING_OCCURRENCE_COUNTER
        case 41: return rn.MISSING_DURATION
        case 42: return rn.MISSING_MAGNITUDE
        case 43: return rn.TIME_THRESHOLD_FOR_UNDER_LIMIT
        case 44: return rn.TIME_THRESHOLD_FOR_OVER_LIMIT
        case 45: return rn.TIME_THRESHOLD_FOR_MISSING_MAGNITUDE
        case 46: return rn.CONTRACTED_VALUE
        case 49: return rn.AVERAGE_VALUE_FOR_RECORDING_INTERVAL_1
        case 50: return rn.AVERAGE_VALUE_FOR_RECORDING_INTERVAL_2
        case 51: return rn.MINIMUM_FOR_RECORDING_INTERVAL_1
        case 52: return rn.MINIMUM_FOR_RECORDING_INTERVAL_2
        case 53: return rn.MAXIMUM_FOR_RECORDING_INTERVAL_1
        case 54: return rn.MAXIMUM_FOR_RECORDING_INTERVAL_2
        case 55: return rn.TEST_AVERAGE
        case 56: return rn.CURRENT_AVERAGE_4_FOR_HARMONICS_MEASUREMENT
        case 58: return rn.TIME_INTEGRAL_4
        case 128: return rn.RU_POSITIVE_DEVIATION
        case 129: return rn.RU_NEGATIVE_DEVIATION
        case 133: return F"{rn.RU_TOTAL_DEVIATION_TIME} {rn.BILLING_PERIOD}"
        case 134: return rn.RU_CHANGE_LIMIT_LEVEL
        case _:  raise ValueError(F'Unknown Processing of measurement values {d}')


def get_rate(value: int) -> str:
    if value == 0:
        return rn.TOTAL
    elif value <= 63:
        return F", {rn.RATE} {value}"
    elif value <= 254:
        return F", {rn.MANUFACTURER_SPECIFIC} {value}"
    elif value == 255:
        return F", {rn.RESERVED}"
    else:
        raise ValueError(F"got group E: {value}, expect 0..255")


def handle_B(value: int) -> str:
    if value == 0:
        return ""
    elif value <= 64:
        return F", {rn.CHANNEL} {value}"
    elif value <= 127:
        return F", {rn.UTILITY_SPECIFIC} {value}"
    elif value <= 199:
        return F", {rn.MANUFACTURER_SPECIFIC} {value}"
    elif value <= 255:
        return F", {rn.RESERVED} {value}"
    else:
        raise ValueError(F"got group B: {value}, expect 0..255")


def handle_E(value: int) -> str:
    if value == 0:
        return ""
    elif value <= 127:
        return F", {rn.INSTANCE} {value}"
    elif value <= 255:
        return F", {rn.MANUFACTURER_SPECIFIC} {value}"
    else:
        raise ValueError(F"got group E: {value}, expect 0..255")


@lru_cache(maxsize=256)
def get_name(logical_name: cst.LogicalName) -> str:
    match logical_name.a, logical_name.b, logical_name.c, logical_name.d, logical_name.e:
        case  0, b, 0, 2, 0:    return F"{rn.ACTIVE_FIRMWARE_IDENTIFIER}{handle_B(b)}"
        case  0, b, 0, 2, 0:    return F"{rn.ACTIVE_FIRMWARE_IDENTIFIER}{handle_B(b)}"
        case  0, b, 0, 2, 1:    return F"{rn.ACTIVE_FIRMWARE_VERSION}{handle_B(b)}"
        case  0, b, 0, 2, 8:    return F"{rn.ACTIVE_FIRMWARE_SIGNATURE}{handle_B(b)}"
        case  0, b, 1, 0, e:    return F"{cn.CLOCK}{handle_B(b)}{handle_E(e)}"
        case  0, b, 2, 0, e:    return F"{cn.MODEM_CONFIGURATION}{handle_B(b)}{handle_E(e)}"
        case  0, b, 10, 0, 0:   return F"{rn.GLOBAL_METER_RESET_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 1:   return F"{rn.MDI_RESET_END_OF_BILLING_PERIOD_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 100: return F"{rn.TARIFFICATION_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 103: return F"{rn.SET_OUTPUT_SIGNALS_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 106: return F"{rn.DISCONNECT_CONTROL_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 107: return F"{rn.IMAGE_ACTIVATION_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 108: return F"{rn.PUSH_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 10, 0, 128: return F"{rn.RU_STOP_FRAME_SCRIPT_TABLE}{handle_B(b)}"
        case  0, b, 11, 0, e:   return F"{cn.SPECIAL_DAYS_TABLE}{handle_B(b)}{handle_E(e)}"
        case  0, b, 12, 0, e:   return F"{cn.SCHEDULE}{handle_B(b)}{handle_E(e)}"
        case  0, b, 13, 0, e:   return F"{cn.ACTIVITY_CALENDAR}{handle_B(b)}{handle_E(e)}"
        case  0, b, 15, 0, 0:   return F"{rn.END_OF_BILLING_PERIOD_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 1:   return F"{rn.DISCONNECT_CONTROL_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 2:   return F"{rn.IMAGE_ACTIVATION_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 3:   return F"{rn.OUTPUT_CONTROL_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 4:   return F"{rn.PUSH_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 5:   return F"{rn.LOAD_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 6:   return F"{rn.M_BUS_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 15, 0, 7:   return F"{rn.FUNCTION_CONTROL_SINGLE_ACTION_SCHEDULE}{handle_B(b)}"
        case  0, b, 16, 1, 1:   return F"{rn.RU_ALARM_MONITOR_1}{handle_B(b)}"
        case  0, b, 17, 0, 0:   return F"{rn.RU_LIMITER_BY_POWER}{handle_B(b)}"
        case  0, b, 17, 0, 1:   return F"{rn.RU_LIMITER_BY_CURRENT}{handle_B(b)}"
        case  0, b, 17, 0, 2:   return F"{rn.RU_LIMITER_BY_VOLTAGE}{handle_B(b)}"
        case  0, b, 17, 0, 3:   return F"{rn.RU_LIMITER_BY_MAGNETIC}{handle_B(b)}"
        case  0, b, 17, 0, 4:   return F"{rn.RU_LIMITER_BY_DIFFERENCE_CURRENT}{handle_B(b)}"
        case  0, b, 17, 0, 5:   return F"{rn.RU_LIMITER_BY_TEMPERATURE}{handle_B(b)}"
        case  0, b, 17, 0, e:   return F"{cn.LIMITER}{handle_B(b)}{handle_E(e)}"
        case  0, 0, 21, 0, 1:   return rn.GENERAL_DISPLAY_READOUT
        case  0, 0, 21, 0, 2:   return rn.ALTERNATE_DISPLAY_READOUT
        case  0, 0, 22, 0, 0:   return rn.RU_IEC_HDLC_SETUP_OPTO
        case  0, 1, 22, 0, 0:   return rn.RU_IEC_HDLC_SETUP_RS_485
        case  0, 2, 22, 0, 0:   return rn.RU_IEC_HDLC_SETUP_GSM
        case  0, b, 22, 0, 0:   return F"{cn.IEC_HDLC_SETUP}{handle_B(b)}"
        case  0, b, 25, 0, 0:   return F"{cn.TCP_UDP_SETUP}{handle_B(b)}"
        case  0, b, 25, 1, 0:   return F"{cn.IPV4_SETUP}{handle_B(b)}"
        case  0, b, 25, 4, 0:   return F"{cn.GPRS_MODEM_SETUP}{handle_B(b)}"
        case  0, b, 25, 6, 0:   return F"{cn.GSM_DIAGNOSTIC}{handle_B(b)}"
        case  0, b, 25, 9, 0:   return F"{cn.PUSH_SETUP}{handle_B(b)}"
        case  0, 0, 40, 0, 0:   return rn.CURRENT_ASSOCIATION
        case  0, 0, 40, 0, 1:   return rn.RU_PUBLIC_CLIENT_ASSOCIATION
        case  0, 0, 40, 0, 2:   return rn.RU_METER_READER_ASSOCIATION
        case  0, 0, 40, 0, 3:   return rn.RU_UTILITY_SETTING_ASSOCIATION
        case  0, 0, 40, 0, 4:   return rn.RU_PUBLIC_CLIENT_ASSOCIATION
        case  0, 0, 40, 0, e:   return F"{cn.ASSOCIATION_LN}{handle_E(e)}"
        case  0, 0, 42, 0, 0:   return rn.COSEM_LOGICAL_DEVICE_NAME
        case  0, 0, 43, 0, e:   return F"{cn.SECURITY_SETUP}{handle_E(e)}"
        case  0, b, 43, 1, e:   return F"{rn.INVOCATION_COUNTER}{handle_B(b)}{handle_E(e)}"
        case  0, 0, 44, 0, e:   return F"{cn.IMAGE_TRANSFER}{handle_E(e)}"
        case  0, 0, 94, 7, 1:   return rn.RU_SPECIFIC_PASSPORT_DATA_PROFILE
        case  0, b, 96, 1, 0:   return F"{rn.RU_DEVICE_FACTORY_NUMBER}{handle_B(b)}"
        case  0, b, 96, 1, 1:   return F"{rn.RU_DEVICE_TYPE}{handle_B(b)}"
        case  0, b, 96, 1, 2:   return F"{rn.RU_DEVICE_METROLOGICAL_VERSION}{handle_B(b)}"
        case  0, b, 96, 1, 3:   return F"{rn.RU_PRODUCER_NAME}{handle_B(b)}"
        case  0, b, 96, 1, 4:   return F"{rn.RU_DEVICE_RELEASE_DATE}{handle_B(b)}"
        case  0, b, 96, 1, 5:   return F"{rn.RU_REMOTE_CONSOLE_SERIAL_NUMBER}{handle_B(b)}"
        case  0, b, 96, 1, 6:   return F"{rn.RU_SPODES_VERSION}{handle_B(b)}"
        case  0, b, 96, 1, 7:   return F"{rn.RU_DEVICE_CONNECTION_SCHEME}{handle_B(b)}"
        case  0, b, 96, 1, 8:   return F"{rn.SPODES3_NOT_METROLOGICAL_FIRMWARE_ID}{handle_B(b)}"
        case  0, b, 96, 1, 9:   return F"{rn.RU_DEVICE_ID}{handle_B(b)}"
        case  0, b, 96, 1, 10:  return F"{rn.RU_COUNTER_POINT_DATA}{handle_B(b)}"
        case  0, b, 96, 2, 0:   return F"{rn.NUMBER_OF_CONFIGURATION_PROGRAM_CHANGES}{handle_B(b)}"
        case  0, b, 96, 2, 1:   return F"{rn.DATE_A_OF_LAST_CONFIGURATION_PROGRAM_CHANGE}{handle_B(b)}"
        case  0, b, 96, 2, 2:   return F"{rn.DATE_A_OF_LAST_TIME_SWITCH_PROGRAM_CHANGE}{handle_B(b)}"
        case  0, b, 96, 2, 3:   return F"{rn.DATE_A_OF_LAST_RIPPLE_CONTROL_RECEIVER_PROGRAM_CHANGE}{handle_B(b)}"
        case  0, b, 96, 2, 4:   return F"{rn.STATUS_OF_SECURITY_SWITCHES}{handle_B(b)}"
        case  0, b, 96, 2, 5:   return F"{rn.DATE_A_OF_LAST_CALIBRATION}{handle_B(b)}"
        case  0, b, 96, 2, 6:   return F"{rn.DATE_A_OF_NEXT_CONFIGURATION_PROGRAM_CHANGE}{handle_B(b)}"
        case  0, b, 96, 2, 7:   return F"{rn.DATE_A_OF_ACTIVATION_OF_THE_PASSIVE_CALENDAR}{handle_B(b)}"
        case  0, b, 96, 2, 10:   return F"{rn.NUMBER_OF_PROTECTED_CONFIGURATION_PROGRAM_CHANGES}{handle_B(b)}"
        case  0, b, 96, 2, 11:   return F"{rn.DATE_A_OF_LAST_PROTECTED_CONFIGURATION_PROGRAM_CHANGE}{handle_B(b)}"
        case  0, b, 96, 2, 12:   return F"{rn.DATE_A_CORRECTED_OF_LAST_CLOCK_SYNCHRONIZATION_SETTING}{handle_B(b)}"
        case  0, b, 96, 2, 13:   return F"{rn.DATE_OF_LAST_FIRMWARE_ACTIVATION}{handle_B(b)}"
        case  0, b, 96, 3, 0:   return F"{rn.I_O_CONTROL_SIGNAL_OBJECTS_GLOBAL}{handle_B(b)}"
        case  0, b, 96, 3, 10:  return F"{cn.DISCONNECT_CONTROL}{handle_B(b)}"
        case  0, b, 96, 3, 20:  return F"{rn.RU_RELAY_LOAD_ARBITRATOR}{handle_B(b)}"
        case  0, 0, 96, 4, 1:   return rn.RU_LCD_BACKLIGHT_MODE
        case  0, 0, 96, 4, 3:   return rn.RU_LOAD_LOCK_STATUS
        case  0, 0, 96, 5, 0:   return rn.INTERNAL_OPERATING_STATUS_GLOBAL
        case  0, 0, 96, 5, 1:   return rn.INTERNAL_OPERATING_STATUS_1
        case  0, 0, 96, 5, 2:   return rn.INTERNAL_OPERATING_STATUS_2
        case  0, 0, 96, 5, 3:   return rn.INTERNAL_OPERATING_STATUS_3
        case  0, 0, 96, 5, 4:   return rn.INTERNAL_OPERATING_STATUS_4
        case  0, 0, 96, 5, 132: return rn.SPODES3_PHASE_ALTERNATING_CONTROL
        case  0, b, 96, 8, 0:   return F"{rn.TIME_OF_OPERATION}{handle_B(b)}"
        case  0, b, 96, 8, 10:  return F"{rn.RU_DURATION_OF_FAILURE_OVERSTRAIN}{handle_B(b)}"
        case  0, b, 96, 9, 0:   return F"{rn.RU_AMBIENT_TEMPERATURE}{handle_B(b)}"
        case  0, b, 96, 11, 0:  return F"{rn.EVENTS_RELATED_TO_VOLTAGE}{handle_B(b)}"
        case  0, b, 96, 11, 1:  return F"{rn.RU_EVENTS_RELATED_TO_CURRENT}{handle_B(b)}"
        case  0, b, 96, 11, 2:  return F"{rn.RU_EVENTS_RELATED_TO_LOAD_RELAY}{handle_B(b)}"
        case  0, b, 96, 11, 3:  return F"{rn.RU_EVENTS_FOR_PROGRAMMING_DEVICE_PARAMETERS}{handle_B(b)}"
        case  0, b, 96, 11, 4:  return F"{rn.RU_EXTERNAL_IMPACT_EVENTS}{handle_B(b)}"
        case  0, b, 96, 11, 5:  return F"{rn.RU_COMMUNICATION_EVENTS}{handle_B(b)}"
        case  0, b, 96, 11, 6:  return F"{rn.RU_ACCESS_CONTROL_EVENTS}{handle_B(b)}"
        case  0, b, 96, 11, 7:  return F"{rn.RU_EVENT_CODES_FOR_THE_SELF_DIAGNOSIS_LOG}{handle_B(b)}"
        case  0, b, 96, 11, 8:  return F"{rn.RU_EVENTS_FOR_EXCEEDING_THE_REACTIVE_POWER}{handle_B(b)}"
        case  0, b, 96, 12, 4:  return F"{rn.RU_CHANNEL_NUMBER_INTERFACE}{handle_B(b)}"
        case  0, b, 96, 12, 6:  return F"{rn.COMMUNICATION_ADDRESS}{handle_B(b)}"
        case  0, 128, 96, 12, 0:  return F"{rn.KPZ_INTERFACE_LIST}"
        case  0, b, 96, 13, 0:  return F"{rn.LOCAL_CONSUMER_MESSAGE}{handle_B(b)}"
        case  0, b, 96, 13, 1:  return F"{rn.DISPLAY_CONSUMER_MESSAGE}{handle_B(b)}"
        case  0, b, 96, 15, 0:  return F"{rn.RU_RELAY_TRIGGERING_METER_FOR_OPENING}{handle_B(b)}"
        case  0, b, 96, 20, 0:  return F"{rn.METER_OPEN_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 1:  return F"{rn.METER_OPEN_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 2:  return F"{rn.METER_OPEN_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 3:  return F"{rn.METER_OPEN_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 5:  return F"{rn.TERMINAL_COVER_OPEN_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 6:  return F"{rn.TERMINAL_COVER_OPEN_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 7:  return F"{rn.TERMINAL_COVER_OPEN_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 8:  return F"{rn.TERMINAL_COVER_OPEN_EVENT_CUMULATIVE}{handle_B(b)}"
        case  0, b, 96, 20, 10:  return F"{rn.TILT_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 11:  return F"{rn.TILT_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 12:  return F"{rn.TILT_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 13:  return F"{rn.TILT_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 15:  return F"{rn.STRONG_DC_MAGNETIC_FIELD_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 16:  return F"{rn.STRONG_DC_MAGNETIC_FIELD_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 17:  return F"{rn.STRONG_DC_MAGNETIC_FIELD_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 18:  return F"{rn.STRONG_DC_MAGNETIC_FIELD_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 20:  return F"{rn.SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 21:  return F"{rn.SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 22:  return F"{rn.SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 23:  return F"{rn.SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 25:  return F"{rn.METROLOGY_TAMPER_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 26:  return F"{rn.METROLOGY_TAMPER_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 27:  return F"{rn.METROLOGY_TAMPER_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 28:  return F"{rn.METROLOGY_TAMPER_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 30:  return F"{rn.COMMUNICATION_TAMPER_EVENT_COUNTER}{handle_B(b)}"
        case  0, b, 96, 20, 31:  return F"{rn.COMMUNICATION_TAMPER_EVENT_TIME_STAMP}{handle_B(b)}"
        case  0, b, 96, 20, 32:  return F"{rn.COMMUNICATION_TAMPER_EVENT_DURATION}{handle_B(b)}"
        case  0, b, 96, 20, 33:  return F"{rn.COMMUNICATION_TAMPER_EVENT_CUMULATIVE_DURATION}{handle_B(b)}"
        case  0, 0, 96, 51, 0:  return rn.RU_BODY_OPENING_STATE
        case  0, 0, 96, 51, 1:  return rn.RU_TERMINALS_COVER_OPENING_STATE
        case  0, 0, 96, 51, 3:  return rn.RU_MAGNETIC_FIELD_STATE
        case  0, 0, 96, 51, 4:  return rn.RU_HF_FIELD_STATE
        case  0, 0, 96, 51, 5:  return rn.RU_ELECTRONIC_SEALS_FIXED_STATE_OF_EVENTS
        case  0, 0, 96, 51, 6:  return rn.RU_PRESSING_ELECTRONIC_SEALS
        case  0, 0, 96, 51, 7:  return rn.RU_CLEAR_OF_ELECTRONIC_SEALS_FIXED_STATE
        case  0, 0, 96, 51, 8:  return rn.RU_FIRST_OPENING_TIME_OF_BODY
        case  0, 0, 96, 51, 9:  return rn.RU_FIRST_OPENING_TIME_OF_TERMINALS_COVER
        case  0, 0, 97, 98, 0:  return rn.RU_ALARM_REGISTER_1
        case  0, 0, 97, 98, 1:  return rn.RU_ALARM_REGISTER_2
        case  0, 0, 97, 98, 10:  return rn.RU_ALARM_FILTER_1
        case  0, 0, 97, 98, 11:  return rn.RU_ALARM_FILTER_2
        case  0, b, 99, 1, e:   return F"{rn.LOAD_PARAMETERS_PROFILE}{handle_B(b)}{handle_E(e)}"
        case  0, b, 99, 13, e:  return F"{rn.GSM_DIAGNOSTIC_PROFILE}{handle_B(b)}{handle_E(e)}"
        case  0, b, 99, 98, 0:  return F"{rn.RU_VOLTAGE_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 1:  return F"{rn.RU_CURRENT_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 2:  return F"{rn.RU_COMMUTATION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 3:  return F"{rn.RU_DATA_CORRECTION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 4:  return F"{rn.RU_EXTERNAL_IMPACT_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 5:  return F"{rn.RU_COMMUNICATION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 6:  return F"{rn.RU_ACCESS_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 7:  return F"{rn.RU_SELF_DIAGNOSTIC_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 8:  return F"{rn.RU_REACTIVE_POWER_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 9:  return F"{rn.RU_QUALITY_POWER_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 10: return F"{rn.RU_STATUS_I_O_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 12: return F"{rn.RU_REACTIVE_POWER_LIMIT_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 13: return F"{rn.RU_TIME_CORRECTION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 14: return F"{rn.RU_START_YEAR_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 15: return F"{rn.RU_QUALITY_FOR_CALCULATION_PERIOD_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 16: return F"{rn.RU_CONTROL_POWER_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 17: return F"{rn.RU_BATTERY_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 18: return F"{rn.RU_CONTROL_OF_LOAD_RELAY_BLOCKER_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 19: return F"{rn.RU_TEMPERATURE_CONTROL_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 20: return F"{rn.RU_VOLTAGE_DEVIATION_LOG} {rn.L1}{handle_B(b)}"
        case  0, b, 99, 98, 21: return F"{rn.RU_VOLTAGE_DEVIATION_LOG} {rn.L2}{handle_B(b)}"
        case  0, b, 99, 98, 22: return F"{rn.RU_VOLTAGE_DEVIATION_LOG} {rn.L3}{handle_B(b)}"
        case  0, b, 99, 98, 23: return F"{rn.RU_LINEAR_VOLTAGE_DEVIATION_LOG} {rn.L1_L2}{handle_B(b)}"
        case  0, b, 99, 98, 24: return F"{rn.RU_LINEAR_VOLTAGE_DEVIATION_LOG} {rn.L2_L3}{handle_B(b)}"
        case  0, b, 99, 98, 25: return F"{rn.RU_LINEAR_VOLTAGE_DEVIATION_LOG} {rn.L3_L1}{handle_B(b)}"
        case  0, b, 99, 98, 26: return F"{rn.RU_OVER_VOLTAGE_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 27: return F"{rn.RU_VOLTAGE_INTERRUPTION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, 28: return F"{rn.RU_ABNORMAL_NETWORK_SITUATION_LOG}{handle_B(b)}"
        case  0, b, 99, 98, e:  return F"{cn.PROFILE_GENERIC}{handle_B(b)}{handle_E(e)}"
        case  0, 0, 128, 100, 0:   return rn.ITE_FIRMWARE_DESCRIPTOR
        case  0, 0, 128, 101, 0:   return rn.ITE_MAGNETIC_SENSOR_STATUS
        case  0, 0, 128, 102, 0:   return rn.ITE_DISCRETE_OUTPUTS
        case  0, 0, 128, 103, 0:   return rn.ITE_SETTING_OF_RELAY_INCLUSION_PER_DAY
        case  0, 0, 128, 150, 0:   return rn.ITE_SETTINGS_MESSAGES
        case  0, 0, 128, 151, 0:   return rn.ITE_CORE_REGISTERS
        case  0, 0, 128, 152, 0:   return rn.ITE_BLE_ID
        case  0, 0, 128, 170, 0:   return rn.ITE_ICCID
        case  0, 0, 199, 255, 255: return cn.CLIENT_SETUP
        case  1, b, 0, 0, e:    return F"{F'{rn.COMPLETE_COMBINED_ELECTRICITY_ID} {e+1}'}{handle_B(b)}"
        case  1, b, 0, 2, 0:    return F"{rn.ACTIVE_FIRMWARE_IDENTIFIER}{handle_B(b)}"
        case  1, b, 0, 2, 8:    return F"{rn.ACTIVE_FIRMWARE_SIGNATURE}{handle_B(b)}"
        case  1, b, 0, 3, 3:    return F"{rn.ACTIVE_ENERGY_OUTPUT_PULSE}{handle_B(b)}"
        case  1, b, 0, 3, 4:    return F"{rn.REACTIVE_ENERGY_OUTPUT_PULSE}{handle_B(b)}"
        case  1, b, 0, 4, 2:    return F"{rn.TRANSFORMER_RATIO_CURRENT}{handle_B(b)}"
        case  1, b, 0, 4, 3:    return F"{rn.TRANSFORMER_RATIO_VOLTAGE}{handle_B(b)}"
        # Nominal values
        case  1, b, 0, 6, 0:    return F"{rn.NOMINAL_VOLTAGE}{handle_B(b)}"
        case  1, b, 0, 6, 1:    return F"{rn.NOMINAL_CURRENT}{handle_B(b)}"
        case  1, b, 0, 6, 2:    return F"{rn.NOMINAL_FREQUENCY}{handle_B(b)}"
        case  1, b, 0, 6, 3:    return F"{rn.MAXIMUM_CURRENT}{handle_B(b)}"
        case  1, b, 0, 6, 4:    return rn.REFERENCE_VOLTAGE_FOR_POWER_QUALITY_MEASUREMENT
        case  1, b, 0, 6, 5:    return F"{rn.REFERENCED_VOLTAGE_FOR_AUX_POWER_SUPPLY}{handle_B(b)}"
        case  1, b, 0, 8, 4:    return F"{rn.RECORDING_INTERVAL_1_FOR_LOAD_PROFILE}{handle_B(b)}"
        case  1, b, 0, 8, 5:    return F"{rn.RECORDING_INTERVAL_2_FOR_LOAD_PROFILE}{handle_B(b)}"
        # Coefficients
        case  1, b, 0, 10, 0:    return F"{rn.TRANSFORMER_MAGNETIC_LOSSES}{handle_B(b)}"
        case  1, b, 0, 10, 1:    return F"{rn.TRANSFORMER_IRON_LOSSES}{handle_B(b)}"
        case  1, b, 0, 10, 2:    return F"{rn.LINE_RESISTANCE_LOSSES}{handle_B(b)}"
        case  1, b, 0, 10, 3:    return F"{rn.LINE_REACTANCE_LOSSES}{handle_B(b)}"
        case  1, b, 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 as c, d, e:
            return F"{handle_B(b)}{rn.CUMULATIVE} {get_obj_names(c)} {get_processing_names(d)} {get_rate(e)}"
        case  1, b, 11 | 12 as c, 7, e:
            return F"{handle_B(b)}{get_obj_names(c)} {get_harmonics_name(e)}"
        case  1, b, 11 | 12 as c, 134, 0:
            return F"{handle_B(b)}{get_obj_names(c)} {rn.RU_CHANGE_LIMIT_LEVEL}"
        case  1, b, 11 | 12 as c, d, e:
            return F"{handle_B(b)}{rn.ANY_PHASE} {get_obj_names(c)} {get_processing_names(d)} {get_harmonics_name(e)}"
        case  1, b, c, d, e if c in range(21, 41):
            return F"{handle_B(b)}{rn.L1} {get_obj_names(c)} {get_processing_names(d)} {get_rate(e)}"
        case  1, b, c, d, e if c in range(41, 61):
            return F"{handle_B(b)}{rn.L2} {get_obj_names(c)} {get_processing_names(d)} {get_rate(e)}"
        case  1, b, c, d, e if c in range(61, 81):
            return F"{handle_B(b)}{rn.L3} {get_obj_names(c)} {get_processing_names(d)} {get_rate(e)}"
        case  1, b, 88 | 89 | 124 | 125 | 126 | 128 | 129 | 130 | 131 | 133 as c, d, e:
            return F"{handle_B(b)}{get_obj_names(c)} {get_processing_names(d)} {get_rate(e)}"
        case  1, b, 91, 7, 131:   return F"{rn.RU_DIFFERENTIAL_CURRENT}. {rn.INSTANTANEOUS_VALUE}{handle_B(b)}"
        case  1, b, 91, 7, 132:   return F"{rn.RU_DIFFERENTIAL_CURRENT}. {rn.PERCENT}. {rn.INSTANTANEOUS_VALUE}{handle_B(b)}"
        case 1, b, 91, d, e if d <= 127:
            return F"{handle_B(b)}{rn.L0_CURRENT_NEUTRAL} {get_processing_names(d)} {get_harmonics_name(e)}"
        case  1, b, 137, d, 128:
            return F"{handle_B(b)}{rn.REACTIVE_FACTOR} {get_processing_names(d)}"
        case  1, 0, 147 as c, 133, 0: return F"{get_obj_names(c)} {rn.BILLING_PERIOD}"
        case  1, 0, 148, 36, 0: return F"{rn.OVER_VOLTAGE_COUNTER} {rn.BILLING_PERIOD}"
        case 1, b, 81, 7, e if e < 78:
            angel_values = (rn.U_L1, rn.U_L2, rn.U_L3, rn.ERROR, rn.I_L1, rn.I_L2, rn.I_L3, rn.I_L0)
            to_, from_ = divmod(logical_name.e, 10)
            return F"{F'{rn.ANGLE_FROM} {angel_values[from_]} {rn.TO} {angel_values[to_]}'}{handle_B(b)}"
        case  1, b, 94, 7, 0:   return F"{rn.RU_PROFILE_OF_CURRENT_VALUES}{handle_B(b)}"
        case  1, b, 94, 7, 1:   return F"{rn.RU_SCALE_PROFILE_FOR_THE_MAGAZINE_OF_MONTHLY_INDICATIONS}{handle_B(b)}"
        case  1, b, 94, 7, 2:   return F"{rn.RU_SCALE_PROFILE_FOR_A_JOURNAL_OF_DAILY_INDICATION}{handle_B(b)}"
        case  1, b, 94, 7, 3:   return F"{rn.RU_SCALE_PROFILE_FOR_CURRENT_FRAMES_OF_CURRENT_VALUES}{handle_B(b)}"
        case  1, b, 94, 7, 4:   return F"{rn.RU_SCALE_PROFILE_FOR_LOAD_PROFILES}{handle_B(b)}"
        case  1, b, 94, 7, 5:   return F"{rn.RU_TELEMECHANICS_PROFILE_FOR_TELEVISION_MEASUREMENTS}{handle_B(b)}"
        case  1, b, 94, 7, 6:   return F"{rn.RU_TELEMECHANICS_PROFILE_OF_TELEVISION_SIGNALING}{handle_B(b)}"
        case  1, b, 94, 7, e:      return F"{F'{rn.COUNTRY_SPECIFIC_IDENTIFIER} #{e+1}'}{handle_B(b)}"
        case  1, b, 98, 1, e:      return F"{rn.RU_MONTHLY_PROFILE}{handle_B(b)}{handle_E(e)}"
        case  1, b, 98, 2, e:      return F"{rn.RU_DAILY_PROFILE}{handle_B(b)}{handle_E(e)}"
        case  1, b, 99, 1, e:      return F"{rn.RU_LOAD_PROFILE}{handle_B(b)}{handle_E(e)}"
        case  1, b, 99, 2, e:      return F"{rn.RU_LOAD_PROFILE} #2{handle_B(b)}{handle_E(e)}"
        case  128, 0, 0, 0, 0:     return rn.ITE_CALIBRATION_STATUS
        case  128, 0, 1, 0, 0:     return rn.ITE_CALIBRATION_APPARENT_POWER
        case  128, 0, 2, 0, 0:     return rn.ITE_CALIBRATION_ACTIVE_POWER
        case  128, 0, 3, 0, 0:     return rn.ITE_CALIBRATION_REACTIVE_POWER
        case  128, 0, 4, 0, 0:     return rn.ITE_CALIBRATION_VOLTAGE
        case  128, 0, 5, 0, 0:     return rn.ITE_CALIBRATION_CURRENT
        case  128, 0, 6, 0, 0:     return rn.ITE_CALIBRATION_ANGLE
        case  128, 0, 7, 0, 0:     return rn.ITE_DISPLAY_SETTING_1
        case  128, 0, 8, 0, 0:     return rn.ITE_DISPLAY_SETTING_2
        case  128, 0, 9, 0, 0:     return rn.ITE_CLOCK_OFFSET_SETTING
        case  128, 0, 10, 0, 0:     return rn.ITE_FACTORY_SETTING_10
        case  128, 0, 11, 0, 0:     return rn.ITE_FACTORY_SETTING_11
        case  128, 0, 12, 0, 0:     return rn.ITE_FACTORY_SETTING_12
        case  128, 0, 13, 0, 0:     return rn.ITE_FACTORY_SETTING_13
        case  128, 0, 14, 0, 0:     return rn.ITE_FACTORY_SETTING_14
        case  128, 0, 15, 0, 0:     return rn.ITE_FACTORY_SETTING_15
        case  128, 0, 16, 0, 0:     return rn.ITE_FACTORY_SETTING_16
        case  128, 0, 17, 0, 0:     return rn.ITE_FACTORY_SETTING_17
        case  128, 0, 18, 0, 0:     return rn.ITE_FACTORY_SETTING_18
        case  128, 0, 19, 0, 0:     return rn.ITE_FACTORY_SETTING_19
        case _:                                    return rn.UNKNOWN
    """ return name by according logical name """



import datetime
import unittest
from src.DLMS_SPODES.types.common_data_types import encode_length
from src.DLMS_SPODES.cosem_interface_classes import ic, collection
from src.DLMS_SPODES.types import cdt, cst, ut, implementations as impl
from src.DLMS_SPODES import relation_to_OBIS, enums
from src.DLMS_SPODES.cosem_interface_classes.collection import Collection


class TestType(unittest.TestCase):
    def test_encode_length(self):
        self.assertEqual(encode_length(1), b'\x01')
        self.assertEqual(encode_length(0x7e), b'\x7e')
        self.assertEqual(encode_length(0x80), b'\x81\x80')
        self.assertEqual(encode_length(0xff), b'\x81\xff')
        self.assertEqual(encode_length(0x100), b'\x82\x01\x00')
        self.assertEqual(encode_length(0x1000), b'\x82\x10\x00')
        self.assertEqual(encode_length(0x10000), b'\x84\x00\x01\x00\x00')
        self.assertEqual(encode_length(0xffffffff), b'\x84\xff\xff\xff\xff')

    def test_exist_attr(self):
        """ Existing attribute 'class_name' in each DLMS class """
        # todo: rewrite, don't work in new API
        # for c_id in ic._COSEM_interface_class_ids:
        #     dlms_class = get_type_from_class(c_id, 0)
        #     self.assertTrue(hasattr(dlms_class, 'NAME'), F'{dlms_class}')

    def test_BitString(self):
        pattern = '101011'
        a = cdt.BitString()
        self.assertEqual(a.encoding, b'\x04\x00', 'default initiation')
        a.set(pattern)
        self.assertEqual(cdt.BitString(pattern), a, 'check set_contents_from')
        self.assertEqual(a.decode(), [1, 0, 1, 0, 1, 1], 'decode to list')

    def test_DateTime(self):
        pattern = datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc)
        self.assertEqual(cdt.DateTime(pattern).decode(), pattern, 'init from datetime and decoding')

    def test_Collection(self):
        collection = Collection()
        collection.add_major(collection.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255')))
        collection.create(class_id=ut.CosemClassId(8), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.1.0.0.255'))
        collection.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.2.29.0.255'))
        self.assertTrue(len(collection) == 3)
        collection.clear()
        self.assertTrue(len(collection) == 1)
        collection.add_if_missing(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        # self.collection.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        self.assertTrue(len(collection) == 1, 'check for not added Association again')

    def test_UnitScaler(self):
        value = cdt.ScalUnitType()
        value.set((10, 10))
        print(value)

    def test_ProfileGeneric(self):
        from src.DLMS_SPODES.cosem_interface_classes.association_ln.ver1 import ObjectListElement
        from src.DLMS_SPODES.types.implementations import structs
        col = Collection()
        col.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        col.create(class_id=ut.CosemClassId(8), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.1.0.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.2.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.1.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.3.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.4.29.0.255'))
        profile = collection.create(class_id=ut.CosemClassId(7), version=cdt.Unsigned(1), logical_name=cst.LogicalName('1.0.94.7.4.255'))
        profile.collection = collection
        profile.set_attr(6, structs.CaptureObjectDefinition().encoding)
        profile.set_attr(3, bytes.fromhex('01 05 02 04 12 00 08 09 06 00 00 01 00 00 ff 0f 02 12 00 00 02 04 12 00 03 09 06 01 00 02 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 01 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 03 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 04 1d 00 ff 0f 03 12 00 00'))
        profile.object_list.selective_access.access_selector.set_contents_from(2)
        profile.set_attr(2, bytes.fromhex('01 01 02 05 09 0c 07 6e 08 1f 07 00 00 ff ff 80 00 00 02 02 0f fd 16 1e 02 02 0f fd 16 1e 02 02 0f fd 16 20 02 02 0f fd 16 20'))
        a = ObjectListElement((3, 0, '1.0.1.29.0.255', None))
        b = col.get_object(a)
        b1 = col.get_object(structs.CaptureObjectDefinition((3, '1.0.1.29.0.255', None, None)))
        # b2 = col.get_object(EntryDescriptor())
        desc1 = profile.get_attr_descriptor(2)
        self.assertEqual(desc1.contents, b'\x00\x07\x01\x00^\x07\x04\xff\x02\x01\x02\x02\x04\x06\x00\x00\x00\x01\x06\x00\x00\x00\x00\x12\x00\x01\x12\x00\x00')
        profile.get_attr(2).selective_access.access_selector.set_contents_from(1)
        desc2 = profile.get_attr_descriptor(2)
        self.assertEqual(desc2.contents, b'\x00\x07\x01\x00^\x07\x04\xff\x02\x01\x01\x02\x04\x02\x04\x12\x00\x01\t\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\t\x0c\x07\xe4'
                                         b'\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff\t\x0c\x07\xe4\x01\x02\xff\xff\xff\xff\xff\x80\x00\xff\x01\x00')
        # e = EntryDescriptor()
        # e.set_contents_from((1, 0, 1, 0))

    def test_Association(self):
        ass = collection.AssociationLNVer0('0.0.40.0.1.255')
        print(ass)

    def test_Conformance(self):
        from src.DLMS_SPODES.cosem_interface_classes.association_ln.ver0 import Conformance, XDLMSContextType
        c = Conformance()
        a = int('1011011101111', 2)
        c.set(a)
        self.assertEqual(c.decode(), [1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        context = XDLMSContextType()
        print(context)

    def test_ImageTransfer(self):
        col = Collection()
        col.create(class_id=ut.CosemClassId(18), version=cdt.Unsigned(0), logical_name=cst.LogicalName('0.0.44.0.0.255'))
        print(col)

    def test_UTF8(self):
        value = cdt.Utf8String()
        value.set('ООО "Курганский приборостроительный завод"')
        print(value)

    def test_Duration(self):
        value = impl.double_long_usingneds.DoubleLongUnsignedSecond()
        value.set(1)
        self.assertEqual(value.report, "0 c")

    def test_ScalerUnitType(self):
        value = cdt.ScalUnitType((1, 4))
        value.unit.set(enums.Unit.CURRENT_AMPERE)
        print(value.unit == cdt.Unit(enums.Unit.CURRENT_AMPERE))



from __future__ import annotations
import datetime
from .__class_init__ import *


class SpecDayEntry(cdt.Structure):
    """ Specifies a special day identifier for a given date. The date may have wildcards for repeating special days like Christmas. """
    values: tuple[cdt.LongUnsigned, cst.OctetStringDate, cdt.Unsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.INDEX, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.SPECIAL_DAY_DATE, cst.OctetStringDate),
                cdt.StructElement(cdt.se.DAY_ID, cdt.Unsigned))  # TODO: must depend with  exec_specdays

    @property
    def index(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def specialday_date(self) -> cst.OctetStringDate:
        """specialday_date formatting follows the rules set in EN 62056-62:2007 4.4.1 for date"""
        return self.values[1]

    @property
    def day_id(self) -> cdt.Unsigned:
        """the range of the day_id must match the length of the bitstring exec_specdays in the related object of interface class “Schedule”"""
        return self.values[2]


class Entries(cdt.Array):
    """ Specifies the list of spec_day_entry """
    TYPE = SpecDayEntry
    values: list[SpecDayEntry]
    __getitem__: SpecDayEntry

    def new_element(self) -> SpecDayEntry:
        indexes: list[int] = [el.index.decode() for el in self.values]
        for i in range(0x1_00_00):
            if i not in indexes:
                return SpecDayEntry((i, None, None))  # TODO: insert first DayID from ActiveCalendar as 3 element
        raise ValueError(F'in {self} all indexes is busy')

    def append_validate(self, element: SpecDayEntry):
        """validate and insert callback for validate change SpecDayEntry"""
        self.__check_index(element.index)
        element.index.register_cb_preset(self.__check_index)

    def __check_index(self, value):
        """validate day_id from DayProfile"""
        if cdt.LongUnsigned(value) in (entry.index for entry in self.values):
            raise ValueError(F'{cdt.LongUnsigned(value)} already exist in {self}')
        else:
            """validate OK"""

    def validate_exist_index(self, value):
        """pass if value in indexes"""
        if cdt.LongUnsigned(value) not in (entry.index for entry in self.values):
            raise ValueError(F'{cdt.LongUnsigned(value)} not exist in {self}')
        else:
            """validate OK"""

    def get_indexes(self) -> list[int]:
        """ getter for callback Index """
        return [entries_element.index.decode() for entries_element in self.values]


class SpecialDaysTable(ic.COSEMInterfaceClasses):
    """ The interface class allows defining dates, which will override normal switching behaviour for special days. The interface class works in
    conjunction with the class "Schedule" or "Activity calendar" and the linking data item is day_id """
    NAME = cn.SPECIAL_DAYS_TABLE
    CLASS_ID = ClassID.SPECIAL_DAYS_TABLE
    VERSION = Version.V0
    A_ELEMENTS = ic.ICAElement(an.ENTRIES, Entries),
    M_ELEMENTS = (ic.ICMElement(mn.INSERT, SpecDayEntry),
                  ic.ICMElement(mn.DELETE, cdt.LongUnsigned))  # Todo: was Delete.with_cb(None, self.entries.get_indexes)

    def characteristics_init(self):
        self.cardinality = (0, 1)

        self._cbs_attr_post_init.update({2: self.__set_delete})
        self.set_attr(2, None)

    @property
    def entries(self) -> Entries:
        return self.get_attr(2)

    @property
    def insert(self) -> SpecDayEntry:
        return self.get_meth(1)

    @property
    def delete(self) -> cdt.LongUnsigned:
        return self.get_meth(2)

    def __set_delete(self):
        try:
            self.delete.register_cb_preset(self.entries.validate_exist_index)
            self.insert.index.register_cb_preset(self.entries.validate_exist_index)
            print('set delete')
        except KeyError:  # At init time
            print('set delete NO:')

    def __delete_entry(self):
        """remove one entry by according delete method index. Call after execute"""
        for entry in self.entries.values:
            if entry.index == self.delete:
                self.entries.values.remove(entry)
                return
        else:
            raise ValueError(F'not found entry with index {self.delete} for remove')

    def get_day_id_of_current_special_day(self, server_time: datetime.datetime = None) -> cdt.Unsigned | None:
        """ return day ID if special day if it in today else None """
        server_time = self.collection.current_time if server_time is None else server_time
        if self.entries is None:
            raise AttributeError('Special days table: attribute Entries is empty. Need receive it from server')
        else:
            current_date: datetime.date = server_time.date()
            for spec_day_entry in self.entries:
                if spec_day_entry.specialday_date.decode() == current_date:
                    return spec_day_entry.day_id
            else:
                return None


if __name__ == '__main__':
    a = SpecialDaysTable('0.0.11.0.0.255')
    print(a)



from __future__ import annotations
from functools import cache
from typing import Callable
from enum import IntFlag, auto
from ... import ITE_exceptions as exc
from ..__class_init__ import *
from ...types import choices
from ...types.implementations.long_unsigneds import ClassId
from ...types.implementations import arrays
from ...pdu_enums import AttributeAccess, MethodAccess


class AccessMode(cdt.Enum):
    """ TODO: """
    ELEMENTS = {b'\x00': en.NO_ACCESS,
                b'\x01': en.READ_ONLY,
                b'\x02': en.WRITE_ONLY,
                b'\x03': en.READ_AND_WRITE}


class AttributeAccessItem(cdt.Structure):
    """ Implemented attribute and it access . Use in Association LN """
    values: tuple[cdt.Integer, AccessMode, cdt.NullData | cdt.Array]
    default = b'\x02\x03\x0f\x01\x16\x00\x00'
    ELEMENTS = (cdt.StructElement(cdt.se.ATTRIBUTE_ID, cdt.Integer),
                cdt.StructElement(cdt.se.ACCESS_MODE, AccessMode),
                cdt.StructElement(cdt.se.ACCESS_SELECTORS, choices.access_selectors))

    @property
    def attribute_id(self) -> cdt.Integer:
        return self.values[0]

    @property
    def access_mode(self) -> AccessMode:
        return self.values[1]

    @property
    def access_selectors(self) -> cdt.NullData | cdt.Array:
        return self.values[2]


class AttributeAccessDescriptor(cdt.Array):
    """ Array of attribute_access_item """
    TYPE = AttributeAccessItem


class MethodAccessItem(cdt.Structure):
    """ Implemented method and it access . Use in Association LN """
    values: tuple[cdt.Integer, cdt.Boolean]
    ELEMENTS = (cdt.StructElement(cdt.se.METHOD_ID, cdt.Integer),
                cdt.StructElement(cdt.se.ACCESS_MODE, cdt.Boolean))

    @property
    def method_id(self) -> cdt.Integer:
        return self.values[0]

    @property
    def access_mode(self) -> cdt.Boolean:
        return self.values[1]


class MethodAccessDescriptor(cdt.Array):
    """ Contain all implemented methods """
    TYPE = MethodAccessItem


class AccessRight(cdt.Structure):
    """ TODO: """
    values: tuple[AttributeAccessDescriptor, MethodAccessDescriptor]
    ELEMENTS = (cdt.StructElement(cdt.se.ATTRIBUTE_ACCESS, AttributeAccessDescriptor),
                cdt.StructElement(cdt.se.METHOD_ACCESS, MethodAccessDescriptor))

    @property
    def attribute_access(self) -> AttributeAccessDescriptor:
        return self.values[0]

    @property
    def method_access(self) -> MethodAccessDescriptor:
        return self.values[1]


class ObjectListElement(cdt.Structure):
    """ Visible COSEM objects with their class_id, version, logical name and the access rights to their attributes and methods within the given application association"""
    values: tuple[cdt.LongUnsigned, cdt.Unsigned, cst.LogicalName, AccessRight]
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.VERSION, cdt.Unsigned),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.ACCESS_RIGHTS, AccessRight))

    @property
    def class_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def version(self) -> cdt.Unsigned:
        return self.values[1]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[2]

    @property
    def access_rights(self) -> AccessRight:
        return self.values[3]


class ObjectListType(arrays.SelectionAccess):
    """ Array of object_list_element. The range for the client_SAP is 0…0x7F. The range for the server_SAP is 0x000…0x3FFF."""
    TYPE = ObjectListElement
    __getitem__: ObjectListElement

    def is_writable(self, ln: cst.LogicalName, indexes: set[int]) -> bool:
        """ index - DLMS object attribute index.
         True: AccessRight is WriteOnly or ReadAndWrite """
        el: ObjectListElement = next(filter(lambda it: it.logical_name == ln, self), None)
        if el is None:
            raise exc.NoObject(F"not find {ln} in object_list")
        item: AttributeAccessItem
        for index in indexes:
            for item in el.access_rights.attribute_access:
                if int(item.attribute_id) == index:
                    if int(item.access_mode) not in (2, 3):
                        return False
                    else:
                        break
                else:
                    continue
            else:
                raise ValueError(F"not find in {ln} attribute index: {index}")
        return True

    def __get_access_right(self, ln: cst.LogicalName | ut.CosemObjectInstanceId) -> AccessRight:
        """return object_list_element of object_list AssociationLN"""
        el: ObjectListElement = next(filter(lambda it: it.logical_name == ln, self), None)
        if el is None:
            raise exc.NoObject(F"not find {ln} in object_list")
        else:
            return el.access_rights

    def get_attr_access(self, ln: cst.LogicalName, index: int) -> AttributeAccess:
        """ index - DLMS object attribute index """
        for item in self.__get_access_right(ln).attribute_access:  # item: AttributeAccessItem
            item: AttributeAccessItem
            if int(item.attribute_id) == index:
                return AttributeAccess(int(item.access_mode))
            else:
                continue
        else:
            raise ValueError(F"not find in {ln} attribute index: {index}")

    def get_meth_access(self, ln: cst.LogicalName | ut.CosemObjectInstanceId, index: int) -> MethodAccess:
        """ index - DLMS object method index """
        for item in self.__get_access_right(ln).method_access:  # item: MethodAccessItem
            item: MethodAccessItem
            if int(item.method_id) == index:
                return MethodAccess(int(item.access_mode))
            else:
                continue
        else:
            raise ValueError(F"not find in {ln} attribute index: {index}")


class ClientSAP(cdt.Enum):
    """ IEC 62056-46 2002 6.4.2.3 Reserved special HDLC addresses p.40. IS15952ver2 """
    TAG = b'\x0f'
    ELEMENTS = {b'\x00': en.NO_STATION,
                b'\x01': en.CLIENT_MANAGEMENT_PROCESS,
                b'\x10': en.PUBLIC_CLIENT,
                b'\x20': en.METER_READER,
                b'\x30': en.UTILITY_SETTING,
                b'\x40': en.PUSH,
                b'\x50': en.FIRMWARE_UPDATE,
                b'\x60': en.IHD}


class ServerSAP(cdt.LongUnsigned):

    def validate(self):
        if int.from_bytes(self.contents, 'big') > 0x3FFF:
            raise ValueError(F'The range for the server_SAP is 0x000…0x3FFF, but got {self.contents.hex()}')


class AssociatedPartnersType(cdt.Structure):
    """ Contains the identifiers of the COSEM client and the COSEM server (logical device) application processes within the physical devices
    hosting these processes, which belong to the application association modelled by the “Association LN” object. """
    values: tuple[ClientSAP, ServerSAP]
    default = (0x10, 1)
    ELEMENTS = (cdt.StructElement(cdt.se.CLIENT_SAP, ClientSAP),
                cdt.StructElement(cdt.se.SERVER_SAP, ServerSAP))

    @property
    def client_SAP(self) -> ClientSAP:
        return self.values[0]

    @property
    def server_SAP(self) -> ServerSAP:
        return self.values[1]


class ApplicationContextName(cdt.Structure):
    """ In the COSEM environment, it is intended that an application context pre-exists and is referenced by its name during the establishment of an
    application association. This attribute contains the name of the application context for that association."""
    values: tuple[cdt.Unsigned, cdt.Unsigned, cdt.LongUnsigned, cdt.Unsigned, cdt.Unsigned, cdt.Unsigned, cdt.Unsigned]
    default = (2, 16, 116, 5, 8, 1, 1)
    ELEMENTS = (cdt.StructElement(cdt.se.JOINT_ISO_CTT_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.COUNTRY_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.COUNTRY_NAME_ELEMENT, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.IDENTIFIED_ORGANIZATION_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.DLMS_UA_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.APPLICATION_CONTEXT_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.CONTEXT_ID_ELEMENT, cdt.Unsigned))

    @property
    def joint_iso_ctt_element(self) -> cdt.Unsigned:
        return self.values[0]

    @property
    def country_element(self) -> cdt.Unsigned:
        return self.values[1]

    @property
    def country_name_element(self) -> cdt.LongUnsigned:
        return self.values[2]

    @property
    def identified_organization_element(self) -> cdt.Unsigned:
        return self.values[3]

    @property
    def DLMS_UA_element(self) -> cdt.Unsigned:
        return self.values[4]

    @property
    def application_context_element(self) -> cdt.Unsigned:
        return self.values[5]

    @property
    def context_id_element(self) -> cdt.Unsigned:
        return self.values[6]


# TODO: join with cdt.FlagMixin
class Conformance(cdt.BitString):
    ELEMENTS = ({Language.ENGLISH: F'reserved-zero', Language.RUSSIAN: F'Зарезервированый-0'},
                {Language.ENGLISH: F'general-protection', Language.RUSSIAN: F'Основной защита'},
                {Language.ENGLISH: F'general-block-transfer', Language.RUSSIAN: F'Основная передача блока'},
                {Language.ENGLISH: F'read', Language.RUSSIAN: F'Чтение'},
                {Language.ENGLISH: F'write', Language.RUSSIAN: F'Запись'},
                {Language.ENGLISH: F'unconfirmed-write', Language.RUSSIAN: F'Неподтвержденная запись'},
                {Language.ENGLISH: F'reserved-six', Language.RUSSIAN: F'Зарезервированый-6'},
                {Language.ENGLISH: F'reserved-seven', Language.RUSSIAN: F'Зарезервированый-7'},
                {Language.ENGLISH: F'attribute0-supported-with-SET', Language.RUSSIAN: F'Поддерживается установка 0 атрибута'},
                {Language.ENGLISH: F'priority-mgmt-supported', Language.RUSSIAN: F'Поддерживается приоритет MGMT'},
                {Language.ENGLISH: F'attribute0-supported-with-GET', Language.RUSSIAN: F'Поддерживается чтение 0 атрибута'},
                {Language.ENGLISH: F'block-transfer-with-get-or-read', Language.RUSSIAN: F'Запрос или чтение через передачу блоков'},
                {Language.ENGLISH: F'block-transfer-with-set-or-write', Language.RUSSIAN: F'Установка или запись через передачу блоков'},
                {Language.ENGLISH: F'block-transfer-with-action', Language.RUSSIAN: F'Активация через передачу блоков'},
                {Language.ENGLISH: F'multiple-references', Language.RUSSIAN: F'Несколько ссылок'},
                {Language.ENGLISH: F'information-report', Language.RUSSIAN: F'Информационный отчет'},
                {Language.ENGLISH: F'data-notification', Language.RUSSIAN: F'Данные-уведомление'},
                {Language.ENGLISH: F'access', Language.RUSSIAN: F'Доступ'},
                {Language.ENGLISH: F'parameterized-access', Language.RUSSIAN: F'Параметризованный доступ'},
                {Language.ENGLISH: F'get', Language.RUSSIAN: F'Извлечение'},
                {Language.ENGLISH: F'set', Language.RUSSIAN: F'Установка'},
                {Language.ENGLISH: F'selective-access', Language.RUSSIAN: F'Выборочный доступ'},
                {Language.ENGLISH: F'event-notification', Language.RUSSIAN: F'Событие-уведомление'},
                {Language.ENGLISH: F'action', Language.RUSSIAN: F'Активация'})
    default = '011000001111111111111011'  # for LN only

    def __init__(self, value: bytes | bytearray | str | int | cdt.BitString = None):
        super(Conformance, self).__init__(value)
        if self.ELEMENTS is not None and len(self) != len(self.ELEMENTS):
            raise ValueError(F'For {self.__class__.__name__} get {len(self)} bits, expected {len(self.ELEMENTS)}')

    def __len__(self):
        return len(self.ELEMENTS)

    def from_bytes(self, value: bytes) -> bytes:
        length, pdu = cdt.get_length_and_pdu(value[1:])
        if length != len(self):
            raise ValueError(F'Got {length=}, expected {len(self)}')
        match value[:1]:
            case self.TAG if len(self) <= len(pdu) * 8: return pdu[:3]
            case self.TAG:                              raise ValueError(F'Got pdu length:{len(pdu)}, expected at least {len(self) >> 3}')
            case _ as error:                            raise TypeError(F'Expected {self.NAME} type, got {cdt.get_common_data_type_from(error).NAME}')

    def from_str(self, value: str) -> bytes:
        value = value + '0' * ((8 - len(self)) % 8)
        list_ = [value[count:(count + 8)] for count in range(0, len(self), 8)]
        value = b''
        for byte in list_:
            value += int(byte, base=2).to_bytes(1, byteorder='little')
        return value

    def from_int(self, value: int) -> bytes:
        if value < 0:
            raise ValueError
        res = 0
        start_bit = 2 ** (len(self) - 1)
        for i in range(len(self)):
            if value & (1 << i):
                res += start_bit >> i
        return res.to_bytes(len(self) // 8, byteorder='big')

    def from_bytearray(self, value: bytearray) -> bytes:
        return bytes(value)

    @classmethod
    def get_values(cls) -> list[str]:
        """ TODO: """
        return [values_dict[get_current_language()] for values_dict in cls.ELEMENTS]

    def validate_from(self, value: str, cursor_position: int) -> tuple[str, int]:
        """ return validated value and cursor position. TODO: copypast FlagMixin """
        type(self)(value=value.zfill(len(self)))
        return value, cursor_position

    @property
    def general_protection(self) -> int:
        return self.decode()[1]

    @property
    def general_block_transfer(self) -> int:
        return self.decode()[2]


class XDLMSContextType(cdt.Structure):
    """ Contains all the necessary information on the xDLMS context for the given association. """
    values: tuple[Conformance, cdt.LongUnsigned, cdt.LongUnsigned, cdt.Unsigned, cdt.Integer, cdt.OctetString]
    default = (None, 1024, 1024, 6, 0, b'\x09\x07\x60\x85\x74\x05\x08\x02\x00')
    ELEMENTS = (cdt.StructElement(cdt.se.CONFORMANCE, Conformance),
                cdt.StructElement(cdt.se.MAX_RECEIVE_PDU_SIZE, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.MAX_SEND_PDU_SIZE, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.DLMS_VERSION_NUMBER, cdt.Unsigned),
                cdt.StructElement(cdt.se.QUALITY_OF_SERVICE, cdt.Integer),
                cdt.StructElement(cdt.se.CYPHERING_INFO, cdt.OctetString))

    @property
    def conformance(self) -> Conformance:
        """the conformance element contains the xDLMS conformance block supported by the server"""
        return self.values[0]

    @property
    def max_receive_pdu_size(self) -> cdt.LongUnsigned:
        """the max_receive_pdu_size element contains the maximum  length for an xDLMS APDU, expressed in bytes that the client may send. This is the same
        as the server-max-receive-pdu-size parameter of the DLMS-Initiate.response pdu (see IEC 62056-53)"""
        return self.values[1]

    @property
    def max_send_pdu_size(self) -> cdt.LongUnsigned:
        """the max_send_pdu_size, in an active association contains the maximum length for an xDLMS APDU, expressed in bytes that the server may send.
        This is the same as the client-max-receive-pdu-size parameter of the DLMS-Initiate.request pdu (see IEC 62056-53)"""
        return self.values[2]

    @property
    def dlms_version_number(self) -> cdt.Unsigned:
        """the dlms_version_number element contains the DLMS version number supported by the server"""
        return self.values[3]

    @property
    def quality_of_service(self) -> cdt.Integer:
        """the quality_of _service element is not used"""
        return self.values[4]

    @property
    def cyphering_info(self) -> cdt.OctetString:
        """the cyphering_info, in an active association, contains the dedicated key parameter of the DLMS-Initiate.request pdu (See 62056-53)"""
        return self.values[5]


class MechanismIdElement(cdt.Enum):
    TAG = b'\x11'
    ELEMENTS = {b'\x00': en.NONE,
                b'\x01': en.LOW,
                b'\x02': en.HIGH,
                b'\x03': en.HIGH_MD5,
                b'\x04': en.HIGH_SHA1,
                b'\x05': en.HIGH_GMAC,
                b'\x06': en.HIGH_SHA256,
                b'\x07': en.HIGH_ECDSA}


class AuthenticationMechanismName(cdt.Structure):
    """ Contains the name of the authentication mechanism for the association (see IEC 62056-53). The authentication mechanism name is specified as an
    OBJECT IDENTIFIER in 7.3.7.2 of IEC 62056-53. The authentication_mechanism_name attribute includes the arc labels of the OBJECT IDENTIFIER. """
    values: tuple[cdt.Unsigned, cdt.Unsigned, cdt.LongUnsigned, cdt.Unsigned, cdt.Unsigned, cdt.Unsigned, MechanismIdElement]
    default = (2, 16, 756, 5, 8, 2, 0)
    ELEMENTS = (cdt.StructElement(cdt.se.JOINT_ISO_CTT_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.COUNTRY_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.COUNTRY_NAME_ELEMENT, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.IDENTIFIED_ORGANIZATION_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.DLMS_UA_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.AUTHENTICATION_MECHANISM_NAME_ELEMENT, cdt.Unsigned),
                cdt.StructElement(cdt.se.MECHANISM_ID_ELEMENT, MechanismIdElement))

    @property
    def joint_iso_ctt_element(self) -> cdt.Unsigned:
        return self.values[0]

    @property
    def country_element(self) -> cdt.Unsigned:
        return self.values[1]

    @property
    def country_name_element(self) -> cdt.LongUnsigned:
        return self.values[2]

    @property
    def identified_organization_element(self) -> cdt.Unsigned:
        return self.values[3]

    @property
    def DLMS_UA_element(self) -> cdt.Unsigned:
        return self.values[4]

    @property
    def authentication_mechanism_name_element(self) -> cdt.Unsigned:
        return self.values[5]

    @property
    def mechanism_id_element(self) -> MechanismIdElement:
        return self.values[6]

    def get_info(self) -> bytes:
        """ info for PDU from application_context_name. """
        return self.joint_iso_ctt_element.contents + \
               self.country_element.contents + \
               self.country_name_element.contents[1:] + \
               self.identified_organization_element.contents + \
               self.DLMS_UA_element.contents + \
               self.authentication_mechanism_name_element.contents + \
               self.mechanism_id_element.contents

    def get_mechanism_id_element(self) -> bytes:
        return self.mechanism_id_element.contents


class AssociationStatus(cdt.Enum):
    """ Enum of access mode for methods """
    ELEMENTS = {b'\x00': en.NON_ASSOCIATED,
                b'\x01': en.ASSOCIATION_PENDING,
                b'\x02': en.ASSOCIATED}


class ClassList(cdt.Array):
    """ Access by class. In this case, only those object_list_elements of the object_list shall be included in the response, which have a class_id
    equal to one of the class_id-s of the class-list. No access_right information is included """
    TYPE = ClassId


class ObjectId(cdt.Structure):
    values: tuple[ClassId, cst.LogicalName]
    default = b'\x02\x02\x12\x00\x08\x09\x06\x00\x00\x01\x00\x00\xff'
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, ClassId),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName))

    @property
    def class_id(self) -> ClassId:
        return self.values[0]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[1]


class ObjectIdList(cdt.Array):
    """ Access by object. The full information record of object instances on the object_Id_list shall be returned. """
    TYPE = ObjectId


class Representation(IntFlag):
    HEX = 0
    ASCII = auto()
    HIDDEN = auto()
    ASCII_HIDDEN = ASCII | HIDDEN


class LLCSecret(cdt.OctetString):
    """ representation for secret """
    __representation: Representation = Representation.HEX
    # used for set class property from instance

    def __init__(self, value: bytes | bytearray | str | int = None):

        super(LLCSecret, self).__init__(value)
        # self.representation = Representation(0)

    def __setattr__(self, key, value):
        match key:
            case 'representation' if not isinstance(value, Representation): raise ValueError(F"Error representation type")
            case 'representation' as rep:                                   LLCSecret.__representation = value
            case _:                                                         super().__setattr__(key, value)

    @property
    def representation(self) -> Representation:
        """used for set class property from instance"""
        return self.__representation

    def __getattr__(self, item):
        raise AttributeError(F'LLCSecret not has {item}')

    @staticmethod
    def __hide_all(value: str) -> str:
        for char_ in filter(lambda it: it != ' ', value):
            value = value.replace(char_, '*')
        return value

    def from_str(self, value: str) -> bytes:
        if self.__representation & Representation.ASCII:
            return cdt.VisibleString.from_str(self, value)
        else:
            return super(LLCSecret, self).from_str(value)

    def __str__(self):
        match self.__representation:
            case Representation.ASCII:                       return cdt.VisibleString.__str__(self)
            case Representation.HIDDEN:                      return self.__hide_all(super(LLCSecret, self).__str__())
            case Representation.ASCII_HIDDEN:                return self.__hide_all(cdt.VisibleString.__str__(self))
            case _:                                          return super(LLCSecret, self).__str__()


class LLCSecretHigh(LLCSecret):
    DEFAULT = b'0000000000000000'

    def validation(self):
        """ check for length equal 16 """
        if len(self) != 16:
            raise ValueError(F'Got length of High secret: {len(self)}, expected 16')

    def validate_from(self, value: str, cursor_position=None) -> tuple[str, int]:
        try:
            correct = type(self)(value)
            return str(correct), cursor_position + (len(str(correct))-len(value))
        except ValueError:
            match self.representation & 0b1:
                case Representation.HEX:
                    cursor_position: int = len(value)-1 if cursor_position is None else cursor_position
                    type(self)(F'{value[:cursor_position]}0{value[cursor_position:]}')  # check possible
                case Representation.ASCII:
                    type(self)(value.zfill(16))
            return value, cursor_position


class AccessSelector(ut.Unsigned8):
    """ Unsigned8 1..4 """
    def __init__(self, value: int | str | ut.Unsigned8 = 1):
        super(AccessSelector, self).__init__(value)
        if int(self) > 4 or int(self) < 1:
            raise ValueError(F'The {self.__class__.__name__} got {self}, expected 1..4')


class Data(ut.Data):
    ELEMENTS = {1: ut.SequenceElement('All information', cdt.NullData),
                2: ut.SequenceElement('Access by class', ClassList),
                3: ut.SequenceElement('Access by object', ObjectIdList),
                4: ut.SequenceElement('Full object information', ObjectId)}


class SelectiveAccessDescriptor(ut.SelectiveAccessDescriptor):
    """ Selective access specification always starts with an access selector, followed by an access-specific access parameter list.
    Specified IS/IEC 62056-53 : 2006, 7.4.1.6 Selective access """
    access_selector: AccessSelector
    access_parameters: Data
    ELEMENTS = (ut.SequenceElement('access_selector', AccessSelector),
                ut.SequenceElement('access_parameters', Data))


class CosemAttributeDescriptorWithSelection(ut.CosemAttributeDescriptorWithSelection):
    access_selection: SelectiveAccessDescriptor
    ELEMENTS = (ut.SequenceElement('cosem_attribute_descriptor', ut.CosemAttributeDescriptor),
                ut.SequenceElement('access_selection', SelectiveAccessDescriptor))


class AssociationLN(ic.COSEMInterfaceClasses):
    """ COSEM logical devices able to establish application associations within a COSEM context using logical name referencing, model the associations
    through instances of the “Association LN” class. A COSEM logical device has one instance of this IC for each association
    the device is able to support"""
    NAME = cn.ASSOCIATION_LN
    CLASS_ID = ClassID.ASSOCIATION_LN_CLASS
    VERSION = Version.V0
    __cb_get_attr_descriptor: Callable[[int], ut.CosemAttributeDescriptor | ut.CosemAttributeDescriptorWithSelection]
    A_ELEMENTS = (ic.ICAElement(an.OBJECT_LIST, ObjectListType, selective_access=SelectiveAccessDescriptor),
                  ic.ICAElement(an.ASSOCIATED_PARTNERS_ID, AssociatedPartnersType),
                  ic.ICAElement(an.APPLICATION_CONTEXT_NAME, ApplicationContextName),
                  ic.ICAElement(an.XDLMS_CONTEXT_INFO, XDLMSContextType),
                  ic.ICAElement(an.AUTHENTICATION_MECHANISM_NAME, AuthenticationMechanismName),
                  ic.ICAElement(an.LLS_SECRET, LLCSecret),
                  ic.ICAElement(an.ASSOCIATION_STATUS, AssociationStatus))
    M_ELEMENTS = (ic.ICMElement(mn.REPLY_TO_HLS_AUTHENTICATION, cdt.OctetString),
                  ic.ICMElement(mn.CHANGE_HLS_SECRET, LLCSecret),
                  ic.ICMElement(mn.ADD_OBJECT, ObjectListElement),
                  ic.ICMElement(mn.REMOVE_OBJECT, ObjectListElement))

    def characteristics_init(self):
        self.set_attr(2, None)
        self.object_list.selective_access = SelectiveAccessDescriptor()
        self.set_attr(3, (0x10*self.logical_name.e, 1) if self.logical_name.e <= 4 else None)
        self.set_attr(4, None)
        self.set_attr(5, None)
        self.set_attr(8, None)
        # init secret after set authentication_mechanism_name(6)
        self._cbs_attr_post_init.update({5: self.__set_dlms_version_to_collection,
                                         6: self.__init_secret,
                                         7: self.__check_mechanism_id_existing})
        # set cb change client SAP
        match self.logical_name.e:
            case 0: self.associated_partners_id.client_SAP.register_cb_preset(self.__handle_preset_current_client_SAP)
            case _: self.associated_partners_id.client_SAP.register_cb_preset(self.__handle_preset_client_SAP)

    @property
    def object_list(self) -> ObjectListType:
        return self.get_attr(2)

    @property
    def associated_partners_id(self) -> AssociatedPartnersType:
        return self.get_attr(3)

    @property
    def application_context_name(self) -> ApplicationContextName:
        return self.get_attr(4)

    @property
    def xDLMS_context_info(self) -> XDLMSContextType:
        return self.get_attr(5)

    @property
    def authentication_mechanism_name(self) -> AuthenticationMechanismName:
        return self.get_attr(6)

    @property
    def LLS_secret(self) -> LLCSecret:
        return self.get_attr(7)

    @property
    def association_status(self) -> AssociationStatus:
        return self.get_attr(8)

    @property
    def reply_to_HLS_authentication(self) -> cdt.OctetString:
        return self.get_meth(1)

    @property
    def change_HLS_secret(self) -> LLS_secret:
        return self.get_meth(2)

    @property
    def add_object(self) -> ObjectListElement:
        return self.get_meth(3)

    @property
    def remove_object(self) -> ObjectListElement:
        return self.get_meth(4)

    def __check_mechanism_id_existing(self):
        """check for existing mechanism ID else ERASE setting"""
        if self.authentication_mechanism_name is None:
            self.clear_attr(7)
            self._cbs_attr_post_init[7] = self.__check_mechanism_id_existing
        else:
            """nothing do it"""

    def __init_secret(self):
        """ before initiating secret need knowledge what kind of mechanism ID """
        match self.authentication_mechanism_name.mechanism_id_element:
            case MechanismIdElement(en.NONE) | MechanismIdElement(en.LOW): self.set_attr_link(7, LLCSecret())
            case MechanismIdElement(en.HIGH):                              self.set_attr_link(7, LLCSecretHigh())
            case unknown:                                                  raise ValueError(F'Not support Secret with {unknown}')

    def __set_dlms_version_to_collection(self):
        self.collection.dlms_ver = int(self.xDLMS_context_info.dlms_version_number)

    @property
    def objects(self) -> list[ic.COSEMInterfaceClasses]:
        """ get all DLMS object in association"""
        try:
            return self.client_objects_list(self.associated_partners_id.client_SAP)
        except AttributeError as e:
            raise exc.NoObject('Objects list is empty')

    @cache
    def client_objects_list(self, value: ClientSAP) -> list[ic.COSEMInterfaceClasses]:
        for association in self.collection.get_objects_by_class_id(ut.CosemClassId(15)):
            if association.associated_partners_id.client_SAP == value and association.logical_name.e != 0:
                match association.object_list:
                    case ObjectListType():
                        ret = list()
                        for obj_list_type in association.object_list:
                            try:
                                obj = self.collection.get_object(obj_list_type)
                                if obj in ret:
                                    print(F'Double intersection {obj}')
                                else:
                                    ret.append(obj)
                            except exc.NoObject as e:
                                print(F'DLMS object not append to client object list. {e}')
                        return ret
                    case _:                raise exc.EmptyObj(F'{association} attr: 2')
        else:
            raise ValueError(F'Not found association with client SAP: {value}')

    def __handle_preset_current_client_SAP(self, value: ClientSAP):
        """ Only for current association. Replacement attributes except logical_name and Client_SAP on attributes of association switching """
        for association in self.collection.get_objects_by_class_id(ut.CosemClassId(15)):
            if association.associated_partners_id.client_SAP == value and association is not self:
                # print('switch to: ', association)  # TODO: for debug
                for index, attr in association.get_index_with_attributes():
                    match index:
                        case 1: continue
                        case 3: self.associated_partners_id.server_SAP.set(attr.server_SAP)
                        case _: self.set_attr_link(index, attr)
                break
        else:
            raise exc.NoObject(F'Not found association with client SAP: {value}')

    def __handle_preset_client_SAP(self, value: cdt.LongUnsigned):
        """ ban change for all associations without current """
        if value != self.associated_partners_id.client_SAP:
            raise ValueError("Don't support change client SAP")

    @property
    def source_address(self) -> bytes:
        """ source address from client_SAP. ISO/IEC 13239:2002(E), Annex H, H.4 Frame format type 3 (page 128). """
        return (int(self.associated_partners_id.client_SAP) << 1 | 1).to_bytes(1, 'big')

    def get_attr_descriptor(self, value: int) -> ut.CosemAttributeDescriptor | CosemAttributeDescriptorWithSelection:
        """ with selection for object_list. TODO: Copypast ProfileGeneric"""
        descriptor: ut.CosemAttributeDescriptor = super(AssociationLN, self).get_attr_descriptor(value)
        if value == 2 and bool(self.collection.current_association.xDLMS_context_info.conformance.decode()[21]):
            return CosemAttributeDescriptorWithSelection((descriptor, self.object_list.selective_access))
        else:
            return descriptor


if __name__ == '__main__':
    as1 = AssociationLN('0.0.40.0.1.255')
    # sela = SelectiveAccessDescriptor()
    # sela.set_selector(3, [(1,2,'1.2.3.4.5.6',None), (4,5,'1.2.3.4.5.4',None),])
    attr_des_with = CosemAttributeDescriptorWithSelection(((15, '0.0.40.0.0.255', 2), (2, [15])))
    a2 = CosemAttributeDescriptorWithSelection(((15, '0.0.40.0.0.255', 2), (4, (15, '0.0.40.0.1.255'))))
    attr_des_with.access_selection.set_selector(2)
    # a = max(map(lambda val: val[Language.RUSSIAN], ClientSAP.elements.values()), key=lambda value: len(value))
    a = Conformance('101010110010101011110101')
    a.set(b'\xb8\x18\x00')
    c = a[1]
    c = a.get_values()
    a = XDLMSContextType()
    a = LLCSecret('11 00 33 34')
    # a = AssociationLN(1)
    a.representation = Representation.ASCII | Representation.HIDDEN
    a1 = LLCSecret('11 55 55')
    print(str(a))
    a = MechanismIdElement(en.HIGH)
    d = MechanismIdElement(en.LOW)
    c = a > d
    a = SelectiveAccessDescriptor()
    b = bytes.fromhex('01 01 02 04 12 00 08 11 00 09 06 00 00 01 00 00 FF 02 02 01 09 02 03 0F 01 16 01 00 02 03 0F 02 16 03 00 02 03 0F 03 16 03 00 02 03 0F 04 16 03 00 02 03 0F 05 16 03 00 02 03 0F 06 16 03 00 02 03 0F 07 16 03 00 02 03 0F 08 16 03 00 02 03 0F 09 16 03 00 01 06 02 02 0F 01 16 01 02 02 0F 02 16 01 02 02 0F 03 16 01 02 02 0F 04 16 01 02 02 0F 05 16 01 02 02 0F 06 16 01')
    a = ObjectListType(b)
    print(a)



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection
from src.DLMS_SPODES.cosem_interface_classes.gsm_diagnostic import SignalQuality


class TestType(unittest.TestCase):
    def test_SignalQuality(self):
        sq = SignalQuality(2)
        self.assertEqual(sq.report, "-109 dBm(2)", "report 2")
        sq.set(30)
        self.assertEqual(sq.report, "-53 dBm(30)", "report 30")



""" COMMON Arrays """
from types import common_data_types as cdt


class Integer(cdt.Array):
    """ Array of Integer """
    TYPE = cdt.Integer


if __name__ == '__main__':
    a = Integer([1,2,3])
    a = Integer(b'\x01\x02\x0f\x0f\x0f\x03')
    a.append('d')
    b = a.decode()
    a.clear()
    print(b)



SCALER = "Масштаб"
UNIT = "Ед. измерения"
DAY_ID = "Идентификатор дня"
DAY_SCHEDULE = "Расписание дня"
START_TIME = "Время старта"
STOP_TIME = "Время окончания"
END_TIME = "Время окончания"
SCRIPT_LOGICAL_NAME = "Логическое имя скрипта"
SCRIPT_SELECTOR = "Селектор скрипта"
SEASON_PROFILE_NAME = "Имя профиля сезона"
SEASON_START = "Начало сезона"
WEEK_NAME = "Имя недели"
WEEK_PROFILE_NAME = "Имя профиля недели"
MONDAY = "понедельник"
TUESDAY = "вторник"
WEDNESDAY = "среда"
THURSDAY = "четверг"
FRIDAY = "пятница"
SATURDAY = "суббота"
SUNDAY = "воскресенье"
ITE_OPTICAL_PORT_NAME = "Имя оптопорта"
ITE_RS485_NAME = "Имя RS-485"
PRESET_TIME = "Заданное время"
VALIDITY_INTERVAL_START = "Начало интервала действия"
VALIDITY_INTERVAL_END = "Конец интервала действия"
PRECEDENCE = "Приоритет"
DELAY = "Задержка"
RELIABILITY = "Надежность"
PEAK_THROUGHPUT = "Пиковая пропускная способность"
MEAN_THROUGHPUT = "Средняя пропускная способность"
DEFAULT = "По умолчанию"
REQUESTED = "По запросу"
CELL_ID = "Идентификатор звонка"
SIGNAL_QUALITY = "Качество сигнала"
LOCATION_ID = "Идентификатор местоположения"
BER = "Ошибка битов"
IMAGE_BLOCK_NUMBER = "Номер блока"
IMAGE_BLOCK_VALUE = "Данные блока"
IMAGE_TO_ACTIVATE_SIZE = "Размер данных"
IMAGE_TO_ACTIVATE_IDENTIFICATION = "Идентификация данных"
IMAGE_TO_ACTIVATE_SIGNATURE = "Подпись данных"
IMAGE_IDENTIFIER = "Идентификатор образа"
IMAGE_SIZE = "размер образа"
ACTION_OVER_THRESHOLD = "Действие выше порога"
ACTION_UNDER_THRESHOLD = "Действие ниже порога"
EMERGENCY_PROFILE_ID = "Идентификатор аварийного профиля"
EMERGENCY_ACTIVATION_TIME = "Аварийное время активации"
EMERGENCY_DURATION = "Продолжительность аварии"
CLASS_ID = "ИК"
LOGICAL_NAME = "Логическое имя объекта"
ATTRIBUTE_INDEX = "Индекс атрибута"
DATA_INDEX = "Индекс данных"
FROM_DATE = "С даты"
TO_DATE = "До даты"
FROM_ENTRY = "С записи"
TO_ENTRY = "До записи"
FROM_SELECTED_VALUE = "От выбранного значения"
TO_SELECTED_VALUE = "До выбранного значения"
RESTRICTING_OBJECT = "Ограничивающий объект"
FROM_VALUE = "От значения"
TO_VALUE = "До значения"
SELECTED_VALUES = "Выбранные объекты"
FIRST_INDEX = "Первый индекс"
LAST_INDEX = "Последний индекс"
FIRST_INDEX_A = "Первый индекс A"
LAST_INDEX_A = "Последний индекс A"
FIRST_INDEX_B = "Первый индекс B"
LAST_INDEX_B = "Последний индекс B"
INDEX = "Индекс"
ENABLE = "Включено"
SWITCH_TIME = "Время переключения"
VALIDITY_WINDOW = "Окно валидности"
EXEC_WEEKDAYS = "Дни недели"
EXEC_SPEC_DAYS = "Спец дни"
BEGIN_DATE = "Дата начала"
END_DATE = "Дата окончания"
SERVICE_ID = "Тип сервиса"
PARAMETER = "Параметр"
SCRIPT_IDENTIFIER = "Идентификатор сценария"
ACTIONS = "Действия"
TIME = "Время"
DATE = "Дата"
SPECIAL_DAY_DATE = "Дата спец дня"
REQUEST = "Запрос"
RESPONSE = "Ответ"
DELAY_AFTER_RESPONSE = "Задержка после ответа"
ATTRIBUTE_ACCESS = "Доступ к атрибуту"
METHOD_ACCESS = "Доступ к методу"
JOINT_ISO_CTT_ELEMENT = "Совместный элемент ISO CTT"
COUNTRY_ELEMENT = "Элемент страны"
COUNTRY_NAME_ELEMENT = "Элемент названия страны"
IDENTIFIED_ORGANIZATION_ELEMENT = "Идентифицированный элемент организации"
DLMS_UA_ELEMENT = "Элемент DLMS UA"
APPLICATION_CONTEXT_ELEMENT = "Элемент контекста приложения"
CONTEXT_ID_ELEMENT = "Элемент идентификатора контекста"
CLIENT_SAP = "клиент SAP"
SERVER_SAP = "сервер SAP"
ATTRIBUTE_ID = "Индекс атрибута"
ACCESS_MODE = "Режим доступа"
ACCESS_SELECTORS = "Селекторы доступа"
AUTHENTICATION_MECHANISM_NAME_ELEMENT = "Элемент контекста приложения"
MECHANISM_ID_ELEMENT = "Идентификационный элемент механизма"
METHOD_ID = "Индекс метода"
VERSION = "Версия"
ACCESS_RIGHTS = "Права доступа"
CONFORMANCE = "Согласование"
MAX_RECEIVE_PDU_SIZE = "Максимальный размер приемного блока pdu"
MAX_SEND_PDU_SIZE = "Максимальный размер отправки pdu"
DLMS_VERSION_NUMBER = "Номер версии DLMS"
QUALITY_OF_SERVICE = "Качество обслуживания"
CYPHERING_INFO = "Информация о шифровании"
KEY_ID = "Идентификатор ключа"
KEY_WRAPPED = "Обертка ключа"
KEY_DATA = "Данные ключа"
CERTIFICATE_ENTITY = "Объект сертификата"
CERTIFICATE_TYPE = "Тип сертификата"
SERIAL_NUMBER = "Серийный номер"
ISSUER = "Издатель"
SUBJECT = "Субъект"
SUBJECT_ALT_NAME = "Альтернативное имя субъекта"
SYSTEM_TITLE = "Системное название"
CERTIFICATE_IDENTIFICATION_TYPE = "Тип идентификации сертификата"
CERTIFICATION_IDENTIFICATION_OPTIONS = "Вариант идентификации сертификата"
REQUEST_ACTOR = "Запрос актуатора"
REQUEST_ACTION_LIST = "Список запросов действий"
ACTION_UP = "Действие вверх"
ACTION_DOWN = "Действие вниз"
ITE_BIT_MAP = "BitMap данные"
RESTRICTION = "Oграничениe"
RESTRICTION_TYPE = "Тип ограничения"
RESTRICTION_VALUE = "Значение ограничения"
COLUMN = "Столбец"
TRANSPORT_SERVICE = "транспортный сервис"
DESTINATION = "адресат"
MESSAGE = "сообщение"
REPETITION_DELAY_MIN = "задержка повтора мин"
REPETITION_DELAY_EXPONENT = "задержка повтора экспонента"
REPETITION_DELAY_MAX = "задержка повтора макс"
CONFIRMATION_START_DATE = "Дата начала подтверждения"
CONFIRMATION_INTERVAL = "Интервал подтверждения"
KEK_ID = "kek идентификатор"
KEY_PARAMETERS = "параметры ключа"
KEY_CIPHERED_DATA = "Ключ зашифрованных данных"
KEY_INFO_TYPE = "тип ключa"
KEY_INFO_OPTIONS = "опции ключа"
KEY_INFO = "информация о ключе"
TRANSACTION_ID = "ID транзакции"
ORIGINATOR_SYSTEM_TITLE = "Название системы создателя"
RECIPIENT_SYSTEM_TITLE = "Название системы получателя"
OTHER_INFORMATION = "Дополнительная информация"
PROTECTION_TYPE = "Тип защиты"
PROTECTION_OPTIONS = "Опции защиты"
